# Copyright 2017 Intel Corporation.
# The source code, information and material ("Material") contained herein is
# owned by Intel Corporation or its suppliers or licensors, and title to such
# Material remains with Intel Corporation or its suppliers or licensors.
# The Material contains proprietary information of Intel or its suppliers and
# licensors. The Material is protected by worldwide copyright laws and treaty
# provisions.
# No part of the Material may be used, copied, reproduced, modified, published,
# uploaded, posted, transmitted, distributed or disclosed in any way without
# Intel's prior express written permission. No license under any patent,
# copyright or other intellectual property rights in the Material is granted to
# or conferred upon you, either expressly, by implication, inducement, estoppel
# or otherwise.
# Any license under such intellectual property rights must be express and
# approved by Intel in writing.


import os
import subprocess
import tempfile
import sys
import functools

# Tensorflow 1.7 has some deprecated features that result in warnings when it is
# imported. Suppress these warnings until TF resolves them.
import warnings
with warnings.catch_warnings():
    warnings.filterwarnings("ignore", category=DeprecationWarning)
    warnings.filterwarnings("ignore", category=FutureWarning)
    import tensorflow as tf

import google.protobuf as proto
from google.protobuf import text_format
import numpy as np
import networkx as nx

from collections import OrderedDict
import math
import re
from tensorflow.core.framework import graph_pb2
from tensorflow.python.framework import ops

from Models.EnumDeclarations import SeedData

from Controllers.EnumController import throw_error, ErrorTable, compiler_assert
from Controllers.Parsers.Parser.Layer import OriginalName, MangledName
from Controllers.Parsers.Parser.Input import Input
from Controllers.Parsers.Parser.Output import Output
from Controllers.Parsers.Parser.DetectionOutput import DetectionOutput
from Controllers.Parsers.Parser.Bias import Bias
from Controllers.Parsers.Parser.Convolution2D import Convolution2D, ConvolutionDepthWise2D

from Controllers.Parsers import TensorFlowParser as tfp
from Controllers.Tensor import UnpopulatedTensor

import Controllers.Preprocess as preprocess
from Controllers.MiscIO import parse_img, preprocess_img
from Controllers.GraphUtils import buildGraph, buildLayerLists
from Controllers import tf_ssd_config_pb2

import skimage.io
import skimage.transform

RAND_HI = 1
RAND_LO = -1

debug = False


# TensorFlow files contain many nodes that are not relevant to inference.  This
# function removes all layers that are not ancecstors of the output layers.
def pruneNodes(parsedLayers, output_node_name):
    outputMangledNames = [layer.getName().stringifyName() for layer in
                        parsedLayers if layer.getName().stringifyOriginalName() == output_node_name]
    g = buildGraph(parsedLayers)
    outputNode = outputMangledNames[0]
    croppedNodes = nx.ancestors(g, outputNode)
    croppedNodes.add(outputNode)
    croppedGraph = nx.subgraph(g, croppedNodes)
    croppedLayers = buildLayerLists(croppedGraph)
    return croppedLayers


def regularizeInPlaceOps(parsedLayers):
    # Some operations in Caffe can be inplace. Introduce new names for these layers
    inPlaceOps = OrderedDict()
    tensorProducer = OrderedDict()
    for layer in parsedLayers:
        for i in layer.getOutputTensorNames():
            try:
                tensorProducer[i.stringifyName()].append(layer)
            except:
                tensorProducer[i.stringifyName()] = [layer]

        if set(layer.getOutputTensorNames()).intersection(set(layer.getInputTensorNames())):
            assert(len(layer.getOutputTensorNames()) == 1)

            tensorName = layer.getOutputTensorNames()[0]
            try:
                inPlaceOps[tensorName.stringifyName()].append(layer)
            except:
                inPlaceOps[tensorName.stringifyName()] = [layer]

    def remangleName(mangledNameList, matchingName):
        for idx, n in enumerate(mangledNameList):
            if n.stringifyName() == matchingName:
                newName = n.remangle()
                mangledNameList[idx] = newName
                return newName

    def replaceName(mangledNameList, matchingName, newName):
        for idx, n in enumerate(mangledNameList):
            if n.stringifyName() == matchingName:
                mangledNameList[idx] = newName

    for tensorName, layerGroup in inPlaceOps.items():
        extendedList = list(set(tensorProducer[tensorName]).difference(set(layerGroup)))
        extendedList.extend(layerGroup[:-1])
        for producer, consumer in zip(extendedList, layerGroup):
            newName = remangleName(producer.getOutputTensorNames(), tensorName)
            replaceName(consumer.getInputTensorNames(), tensorName, newName)

    return parsedLayers

def createTensors(parsedLayers):
    # Collect all the tensorNames and sizes
    tensorNames = OrderedDict()
    for layer in parsedLayers:
        for tensorName, tensorSize in zip(layer.getInputTensorNames(), layer.getInputTensorSizes()):
            if tensorName.stringifyName() not in tensorNames:
                tensorNames[tensorName.stringifyName()] = UnpopulatedTensor(tensorSize)
                tensorNames[tensorName.stringifyName()].setName(tensorName)

        for tensorName, tensorSize in zip(layer.getOutputTensorNames(), layer.getOutputTensorSizes()):
            if tensorName.stringifyName() not in tensorNames:
                tensorNames[tensorName.stringifyName()] = UnpopulatedTensor(tensorSize)
                tensorNames[tensorName.stringifyName()].setName(tensorName)

    for layer in parsedLayers:
        layer.setInputTensors([tensorNames[n.stringifyName()] for n in layer.getInputTensorNames()])
        layer.setOutputTensors([tensorNames[n.stringifyName()] for n in layer.getOutputTensorNames()])

def insertOutputOps(parsedLayers, output_name):
    # Find all tensors that are not consumed by anybody
    tensorNames = OrderedDict()
    for layer in parsedLayers:
        for tensor in layer.getInputTensors():
            if tensor.getName().stringifyName() not in tensorNames:
                tensorNames[tensor.getName().stringifyName()] = ([0, 0], tensor, [])

            tensorNames[tensor.getName().stringifyName()][0][0] += 1

        for tensor in layer.getOutputTensors():
            if tensor.getName().stringifyName() not in tensorNames:
                tensorNames[tensor.getName().stringifyName()] = ([0, 0], tensor, [])

            tensorNames[tensor.getName().stringifyName()][0][1] += 1
            tensorNames[tensor.getName().stringifyName()][2].append(layer)

    for tensorName, tensorValue in tensorNames.items():
        consumersAndProducers, tensor, producers = tensorValue
        consumers = consumersAndProducers[0]
        if consumers == 0 and tensor.getName().stringifyOriginalName() == output_name:
            x = Output('output', [tensor.getName()], [])
            x.setInputTensors([tensor])
            x.setOutputTensors([])
            x.loadInputTensorSizes([tensor.getShape()])
            x.loadOutputTensorSizes([])

            assert(len(producers) == 1)
            if isinstance(producers[0], DetectionOutput):
                x.enableDetectionOutput()

            parsedLayers.append(x)

    return parsedLayers

def printLayers(layers):
    for layer in layers:
        print('Node:', layer)
        print('  layer name:', layer.getStringifiedName())
        print('  input tensors:', )
        for tensorName in layer.getInputTensorNames():
            print('    ', tensorName.stringifyName())
        print('  output tensors:')
        for tensorName in layer.getOutputTensorNames():
            print('    ', tensorName.stringifyName())


def getDirectAncestorOps(ops):
    directAncestorOps = []
    for o in ops:
        directAncestorOps += [i.op for i in o.inputs]
    return directAncestorOps

def opHasInputAsAncestor(op, input_op, parsedLayers):
    # Checks if an op node has the input_op as

    if (op == input_op):
        return True

    # 1. check if any of the direct ancestors is already parsed
    parsedOpsNames = [p.getStringifiedOriginalName() for p in parsedLayers]
    iterate = True
    parentOps = [op]
    while len(parentOps) > 0:
        parentOps = getDirectAncestorOps(parentOps)
        parentOpsNames = [p.name for p in parentOps]

        if any(pName in parsedOpsNames for pName in parentOpsNames):
            # Parent is already parsed, hence the operation is non-constant
            return True

    return False

class TensorFlowParser:

    def __init__(self):
        self.type = 'TensorFlow'
        self.model = None
        self.feed_dict = None

        self.session = None
        self.tf_input_op = None
        self.myriad_input_op = None
        self.tf_output_ops = None
        self.myriad_output_op = None
        self.ssd_config = None
        self.input_shape = None

    def getType(self):
        return self.type

    def load_ssd_config(self, path):
        with open(path, "r") as config_pb:
            self.ssd_config = tf_ssd_config_pb2.TFSSDConfig()
            text_format.Parse(config_pb.read(), self.ssd_config)

    def get_ssd_config(self):
        return self.ssd_config

    def get_mbox_loc_conf_tensor_names(self):
        """ In TensorFlow the mbox_loc and mbox_conf tensors are generated by
            the BoxPredictor blocks and packed via the 2 concat layers named:
                - concat for mbox_loc
                - concat_1 for mbox_conf

            The function tries to find these 2 concat layers and extract tehir input
            tensor names in order to pass them further to the insert_ssd_patch_layers
            module which will complet the network.
        """
        try: mbox_loc_concat = self.model.get_operation_by_name("concat")
        except:
            throw_error(ErrorTable.NoOutputNode,
                        "concat. (This might not be an SSD network)")

        try:
            mbox_conf_concat = self.model.get_operation_by_name("concat_1")
        except:
            throw_error(ErrorTable.NoOutputNode,
                        "concat_1. (This might not be an SSD network)")

        mbox_loc_tensor_names = [tensor.name for tensor in mbox_loc_concat.inputs
                                 if tensor.op.type != "Const"]
        mbox_conf_tensor_names = [tensor.name for tensor in mbox_conf_concat.inputs
                                 if tensor.op.type != "Const"]

        return mbox_loc_tensor_names, mbox_conf_tensor_names

    def get_prior_boxes(self, arguments):
        """ In TensorFow the PriorBox layers are replaced by a more generic module
            called MultipleGridAnchorGenerator. Because the equivalation between the
            MultipleGridAnchorGenerator module and PriorBox layers is not straight
            forward and other problems with isertion of PriorBox layers into the graph
            we need to evaluate the PriorBoxes directly from tensorflow at compile time.

            Returns: prior_boxes - numpy array containing the box priors.
        """

        # The TF SSD graphs present 2 diffrent structures with respect to
        # MultipleGridAnchorGenerator. The diffrence between the 2 structures is
        # where the concatenation of the generated boxes is done and it can be done
        # inside or outside the MultipleGridAnchorGenerator scope.
        # We must search for the layer/op that is connected to the
        # MultipleGridAnchorGenerator scope block. If that layer is a concat it means
        # that the concatenation is done "outside" and we need to evaluate the output
        # of that concat operation. If the layer is something diffrent from a concat
        # it means that the concatenation is done "inside" and we need to evalutate
        # the input tensor of that operation.

        # Prune operations by scope to have a smaller search space.
        operations = [op for op in self.model.get_operations()
                      if not (re.search("FeatureExtractor", op.name) or
                              re.search("BoxPredictor", op.name) or
                              re.search("MultipleGridAnchorGenerator", op.name))
                      ]

        prior_box_fetches = None
        for op in operations:
            for input_tensor in op.inputs:
                if re.search("MultipleGridAnchorGenerator", input_tensor.name):
                    if op.type in ["Concat", "ConcatV2"]:
                        prior_box_fetches = op.outputs[0]
                    else:
                        prior_box_fetches = op.inputs[0]

        if prior_box_fetches is None:
            # If the TensorFlow model is optimized the MultipleGridAnchorGenerator
            # output(i.e. prior boxes) is computed and introduced as a Const
            # operation in the Preprocessor block that will have control inputs
            # from MultipleGridAnchorGenerator. Try to find this const tensor.
            for op in operations:
                for control_input in op.control_inputs:
                    if re.search("MultipleGridAnchorGenerator", control_input.name):
                        if op.type in ["Const"]:
                            prior_box_fetches = op.outputs[0]

        assert prior_box_fetches is not None, "Could not find priorboxes."

        feed_dict = {self.tf_input_op.outputs[0] : np.ones(self.get_input_shape(arguments))}
        prior_boxes = self.session.run(prior_box_fetches, feed_dict)

        prior_boxes = np.squeeze(prior_boxes)

        assert len(prior_boxes.shape) == 2, "Wrong shape for the PriorBoxes."

        return prior_boxes


    def _get_input_ops(self, arguments):
        """ Gets the input operations for both TF reference and myriad network.
            The input tensors for the TF network might differ from the myriad network.
            This is the case for the TensorFlow SSD networks where the input images
            if passed through a Preprocessing block and the myriad input is the input
            to the FeatureExtractor block.

            Returns: (tf_input_op, myriad_input_op)
        """
        if self.ssd_config is not None and arguments.input_node_name is not None:
            throw_error(ErrorTable.NoOutputNode, ". Input node flag not supported for SSD networks.")

        if self.tf_input_op is not None and self.myriad_input_op is not None:
            return (self.tf_input_op, self.myriad_input_op)

        assert self.session is not None, "Parser is not initialized."
        with self.session.as_default():
            if arguments.input_node_name is not None:
                try:
                    self.tf_input_op = self.model.get_operation_by_name(arguments.input_node_name)
                    self.myriad_input_op = self.tf_input_op
                except:
                    throw_error(ErrorTable.NoOutputNode, arguments.input_node_name)
            else:
                # Search for Placeholder operations. If there is only one Placeholder
                # select that as input.
                place_holder_ops = [
                    op for op in self.model.get_operations() if op.type == "Placeholder"
                    ]
                if len(place_holder_ops) != 1:
                    throw_error(ErrorTable.NoOutputNode, ". Could not infer input node name.")

                self.tf_input_op = place_holder_ops[0]
                if self.ssd_config is None:
                    self.myriad_input_op = self.tf_input_op
                else:
                    # For TF SSD networks we ar interestead in the FeatureExtractor
                    # and BoxPredictor_n blocks.
                    # The input to the myriad network is going to be the layer from
                    # the Preprocessor block that is connected to FeatureExtractor
                    # block
                    for op in self.model.get_operations():
                        if op.name.find("FeatureExtractor") != -1:
                            for input_tensor in op.inputs:
                                if input_tensor.name.find("Preprocessor") != -1:
                                    feat_ext_input_t = input_tensor

                    for op in self.model.get_operations():
                        if op.name.find("Preprocessor") != -1:
                            if feat_ext_input_t in op.outputs:
                                self.myriad_input_op = op
                                break

        return (self.tf_input_op, self.myriad_input_op)


    def _get_parent_ops(self, op):
        """ Gets all parent ops for the op sent as parameter.

            Returns: [parent_ops]
        """
        parent_ops = [i.op for i in op.inputs]
        return parent_ops


    def _get_output_ops(self, arguments):
        """ Gets the output operations for both TF reference and myriad network.
            The output operations for the TF network might differ from the myriad network.
            This is the case for the TensorFlow SSD networks where there are 4 output
            operations: num_detections, detection_scores, detection_boxes and detection_classes.

            Returns: ([tf_output_ops], [myriad_output_op]).
                     For SSD networks returns:
                     ([num_detections_op, detection_scores_op, detection_classes_op,
                       detection_boxes_op], None)
        """

        if self.tf_output_ops is not None and self.myriad_output_op is not None:
            return (self.tf_output_ops, self.myriad_output_op)

        assert self.session is not None, "Parser is not initialized."
        with self.session.as_default():
            if arguments.output_node_name is not None:
                if self.ssd_config is not None:
                    throw_error(ErrorTable.NoOutputNode,
                                ". Output node flag not supported for SSD networks.")
                else:
                    try:
                        self.tf_output_ops = [self.model.get_operation_by_name(arguments.output_node_name)]
                        self.myriad_output_op = self.tf_output_ops
                    except:
                        throw_error(ErrorTable.NoOutputNode, arguments.output_node_name)
            else:
                if self.ssd_config is not None:
                    inferred_output_names = ["num_detections", "detection_scores",
                                             "detection_classes", "detection_boxes"]
                else:
                    inferred_output_names = ["output"]

                #Search for the inferred output operations by name.
                try:
                    self.tf_output_ops = [self.model.get_operation_by_name(op_name)
                                          for op_name in inferred_output_names]
                except:
                    throw_error(ErrorTable.NoOutputNode,
                                ". Could not infer output node names. Default names:" +
                                functools.reduce(lambda s0, s1: s0 + "," + s1, inferred_output_names))

                if self.ssd_config is None:
                    self.myriad_output_op = self.tf_output_ops

        return (self.tf_output_ops, self.myriad_output_op)

    def get_input_shape(self, arguments):
        """ Infer input shape for the TF model.
        """

        if self.input_shape is not None:
            return self.input_shape

        tf_input_op, _ = self._get_input_ops(arguments)
        inferred_input_shape = tf_input_op.outputs[0].get_shape().as_list()

        if self.ssd_config is not None:
            ssd_pipline_shape = (1, self.ssd_config.input_height, self.ssd_config.input_width, 3)

            for axis in range(0, 4):
                if inferred_input_shape[axis] != None:
                    assert inferred_input_shape[axis] == ssd_pipline_shape[axis], \
                           "SSD pipline shape does not match model expected shape" + \
                           "model shape {0} != ".format(inferred_input_shape) + \
                           "pipline config shape {0}".format(ssd_pipline_shape)


            self.input_shape = ssd_pipline_shape
        else:
            if inferred_input_shape[0] == None:
                # Batch size can be runtime defined (i.e. None). If so set it to 1
                # since we only support batch sizes of 1.
                inferred_input_shape[0] = 1
            elif inferred_input_shape[0] != 1:
                throw_error(ErrorTable.StageDetailsNotSupported,
                            "Input tensor batch size != 1, not supported.")

            for axis in inferred_input_shape:
                if axis is None:
                    throw_error(ErrorTable.StageDetailsNotSupported,
                                "Dynamic input shape not supported.")

            self.input_shape = inferred_input_shape

        return self.input_shape

    def calculateReference(self, arguments):
        # Input shape is in NHWC
        input_shape = self.get_input_shape(arguments)
        nhwc_to_nchw_transpozition = (0, 3, 1, 2)

        if isinstance(arguments.image, SeedData):
            raw_input_data = preprocess.gen_random_image(arguments.image, arguments.seed, input_shape)
            processed_input_data = raw_input_data
        else:
            raw_input_data = preprocess.input_to_npy(arguments.image, input_shape)
            mean = preprocess.load_mean(arguments.mean)
            scale = preprocess.load_scale(arguments.raw_scale)
            channel_swap = preprocess.load_channel_swap(arguments.channel_swap)
            if self.ssd_config is None:
                raw_input_data = preprocess.resize(raw_input_data, input_shape[1], input_shape[2])
            processed_input_data = preprocess.preprocess_image(raw_input_data, mean, scale, channel_swap)
            processed_input_data = preprocess.match_network_shape(processed_input_data,
                                                                  [input_shape[0],
                                                                   input_shape[3],
                                                                   input_shape[1],
                                                                   input_shape[2]])

            raw_input_data = preprocess.match_network_shape(raw_input_data,
                                                            [input_shape[0],
                                                             input_shape[3],
                                                             input_shape[1],
                                                             input_shape[2]])

        tf_input_op, _ = self._get_input_ops(arguments)
        tf_output_ops, _ = self._get_output_ops(arguments)

        with self.session.as_default():
            tf.global_variables_initializer()

            output_fetches = {op.name: op.outputs[0] for op in tf_output_ops}
            tf_input = processed_input_data if self.ssd_config is None else raw_input_data
            feed_dict = {tf_input_op.outputs[0] : tf_input}
            self.feed_dict = feed_dict

            output_fetches = self.session.run(output_fetches, feed_dict)

        if self.ssd_config is not None:
            num_detections = int(output_fetches["num_detections"][0])
            expected_result = np.zeros((1, 1, num_detections, 7))
            expected_result[:, :, :, 1] = output_fetches["detection_classes"][0, 0:num_detections]
            expected_result[:, :, :, 2] = output_fetches["detection_scores"][0, 0:num_detections]

            expected_result[:, :, :, 3] = output_fetches["detection_boxes"][0, 0:num_detections, 1]
            expected_result[:, :, :, 4] = output_fetches["detection_boxes"][0, 0:num_detections, 0]
            expected_result[:, :, :, 5] = output_fetches["detection_boxes"][0, 0:num_detections, 3]
            expected_result[:, :, :, 6] = output_fetches["detection_boxes"][0, 0:num_detections, 2]
        else:
            assert len(tf_output_ops) == 1, "Unexpected number of outputs"
            expected_result = output_fetches[tf_output_ops[0].name]

            extend_shape = lambda x: np.insert(x, 0, [1]*max((4 - len(x)), 0))
            expected_result = np.transpose(
                                expected_result.reshape(extend_shape(expected_result.shape)),
                                nhwc_to_nchw_transpozition)

        myriad_input_data = np.transpose(
                                processed_input_data.astype(dtype=np.float16),
                                nhwc_to_nchw_transpozition)

        return myriad_input_data, expected_result


    def loadNetworkObjects(self,graph_path, model_path=None):
        """ Get the tensorflow protobuff model and parse it via tensorflow
        """
        print(graph_path)

        self.session = tf.Session()
        with self.session.as_default():
            filetype = graph_path.split(".")[-1]
            if filetype == 'pb':
                graph_def = graph_pb2.GraphDef()
                with open(graph_path, 'rb') as f:
                    graph_def.ParseFromString(f.read())
                tf.import_graph_def(graph_def, name="")
            else:
                saver = tf.train.import_meta_graph(graph_path, clear_devices=True)
                if saver is not None:
                    saver.restore(self.session, graph_path[:graph_path.rfind('.')])

            self.model = tf.get_default_graph()

        return

    def _gen_dummy_feed_dict(self, arguments):
        """
            Generate a dummy feed_dict with the correct input shape
        """
        feed_dict = {self.tf_input_op.outputs[0] : np.ones(self.get_input_shape(arguments))}
        return feed_dict

    def _load_tensor_sizes(self, parsed_layers, arguments):
        # Gen dummy feed dict with correct input shape.
        feed_dict = {self.tf_input_op.outputs[0] : np.ones(self.get_input_shape(arguments))}

        tensor_shape_fetches = {tensor_name : tf.shape(self.model.get_tensor_by_name(tensor_name))
                                for layer in parsed_layers for tensor_name in
                                layer.getInputTensorNames() + layer.getOutputTensorNames()}

        tensor_shape_fetches = self.session.run(tensor_shape_fetches, feed_dict)

        nhwc_to_nchw_transpozition = [0, 3, 1, 2]
        extend_shape = lambda x: np.insert(x, 0, [1]*max((4 - len(x)), 0))
        tensor_shape_fetches = {
            name : extend_shape(shape)[nhwc_to_nchw_transpozition].tolist()
            for name, shape in tensor_shape_fetches.items()}

        for layer in parsed_layers:
            layer.loadTensorSizes(tensor_shape_fetches)

        return parsed_layers

    def parse(self, arguments):
        # Define which subparser needs to be called for each layer
        subParsers = {
            'Placeholder': tfp.Placeholder.load,
            'Conv2D': tfp.Convolution.load,
            'DepthwiseConv2dNative': tfp.Convolution.load,
            'MaxPool': tfp.Pooling.load,
            'AvgPool': tfp.Pooling.load,
            'Relu': tfp.ReLU.load,
            'Relu6': tfp.ReLU.load,
            'BiasAdd': tfp.BiasAdd.load,
            'Concat': tfp.Concat.load,
            'ConcatV2': tfp.Concat.load,
            'Identity' : tfp.Identity.load,
            'Slice': tfp.Slice.load,
            'Add': tfp.Math.load,
            'Mul': tfp.Math.load,
            'Sub': tfp.Math.load,
            'LRN': tfp.LRN.load,
            'MatMul': tfp.MatMul.load,
            'Softmax': tfp.Softmax.load,
            'Sigmoid': tfp.Sigmoid.load,
            'FusedBatchNorm': tfp.FusedBatchNorm.load,
            # TODO: Needs to be checked first.
            'Reshape' : tfp.Reshape.load,
            'Squeeze' : tfp.Identity.load,
            'Pad' : tfp.Pad.load,
            'Maximum' : tfp.Math.load,
            'RealDiv' : tfp.RealDiv.load,
            'Sum' : tfp.Sum.load,
            'Square' : tfp.Square.load,
            'Rsqrt' : tfp.Rsqrt.load,
            'Conv2DBackpropInput' : tfp.Conv2DBackpropInput.load,
            'Mean' : tfp.Mean.load,
            'Split' : tfp.Split.load,
            'Const' : "skip",
            'RandomUniform' : "skip",
            'Variable' : "skip",
            'VariableV2' : "skip",
            'Assign' : "skip",
            'Pack' : "skip",
            'Shape' : "skip",
            'StridedSlice' : "skip",
        }

        tf_input_op, myriad_input_op = self._get_input_ops(arguments)
        tf_output_ops, myriad_output_op = self._get_output_ops(arguments)

        with self.session.as_default():
            parsedLayers = [Input(myriad_input_op.name, [], [myriad_input_op.outputs[0].name])]

            operations = self.model.get_operations()
            if self.ssd_config is not None:
                operations = [op for op in operations
                              if (op.name.find("FeatureExtractor") != -1 or
                                  op.name.find("BoxPredictor") != -1)
                              ]

            op_parser_dict = {myriad_input_op: parsedLayers[0]}
            dummy_feed_dict = self._gen_dummy_feed_dict(arguments)
            mean_ops = []
            for op in operations:
                if op == myriad_input_op:
                    continue

                opParser = subParsers.get(op.type, None)
                if opParser == "skip" or \
                    all(op_parser_dict[p] == "skip" for p in self._get_parent_ops(op)):
                    op_parser_dict[op] = "skip"
                    continue
                elif opHasInputAsAncestor(op, myriad_input_op, parsedLayers):
                    compiler_assert(opParser is not None,
                                    ErrorTable.StageTypeNotSupported, op.type)
                    parsed_op = opParser(op, operations, dummy_feed_dict)
                    if op.type == 'Mean':
                        mean_ops.extend(parsed_op)
                    parsedLayers.extend(parsed_op)
                    op_parser_dict[op] = opParser
                else:
                    op_parser_dict[op] = "skip"

                if self.ssd_config is None and op == myriad_output_op[0]:
                    break

            self._load_tensor_sizes(parsedLayers, arguments)

            # The Mean operations are translated into global AveragePool
            # operations for which the kernel size is equal to the input size.
            # Because the input size might be dynamic we need to set the kernel
            # size for the Mean operations after the tensor sizes have been loaded.
            for mean_op in mean_ops:
                input_size = mean_op.getInputTensorSizes()
                mean_op.loadKernelSize(input_size[0][2], input_size[0][3])

            # Mangle tensor names
            tensorNamesDict = OrderedDict()
            for layer in parsedLayers:
                for tensorName in list(layer.getInputTensorNames()) + list(layer.getOutputTensorNames()):
                    if tensorName not in tensorNamesDict:
                        tensorNamesDict[tensorName] = MangledName(OriginalName(tensorName))

            # Replace tensor names in layers with mangled ones:
            for layer in parsedLayers:
                layer.setInputTensorNames([tensorNamesDict[name] for name in layer.getInputTensorNames()])
                layer.setOutputTensorNames([tensorNamesDict[name] for name in layer.getOutputTensorNames()])

            # Convert inPlace operations into regular ones
            parsedLayers = regularizeInPlaceOps(parsedLayers)

            # Prune subgraphs
            # Is this still necessary after implementing the method with checking
            # for constant OP at parsing time ?
            if self.ssd_config is None:
                parsedLayers = pruneNodes(parsedLayers, myriad_output_op[0].name)

            # Create tensor objects for each operation
            createTensors(parsedLayers)

            # Insert Output operation for non SSD networks. The output of SSD network
            # is going to be added in insert_ssd_patch_layers Phase.
            if self.ssd_config is None:
                parsedLayers = insertOutputOps(parsedLayers,
                                               myriad_output_op[0].outputs[0].name)

            g = buildGraph(parsedLayers)

            print("Fusing DeptwiseConv + Pointwise Convolution into plain Convolution")
            g = DwSeparableConvToConv2D(g)
            print("Fusing Add and Batch after Convolution")
            g = fuseBiasAdd(g)

            parsedLayers = buildLayerLists(g)

        return parsedLayers

    def get_layers_data(self, outBlobNames):

        with self.session.as_default():
            tensors = [self.model.get_tensor_by_name(name) for name in outBlobNames]
            out_data = self.session.run(tensors, self.feed_dict)

            for idx, t in enumerate(out_data):
                if (len(t.shape) == 4):
                    out_data[idx] = t.transpose((0, 3, 1, 2))

        return out_data


def DwSeparableConvToConv2D(g):
    from Controllers.Optimizer import fuse_nodes
    from Models.EnumDeclarations import PadStyle

    """
        Iterates over the graph detecting sequences of DeptwiseConv2D followed by
        PointwiseConvolution (i.e. 1x1 Convolution), sequence known as
        Deptwise Separable Convolution, into a standard Convolution2D
    """
    def isDepthwiseConv(layer):
        """
            Returns True/False if the given layer is of type ConvolutionDepthWise2D
            and its parameters are qualified for the fusing of separable convolution
        """
        from Controllers.Parsers.Parser.Convolution2D import ConvolutionDepthWise2D

        if type(layer) not in [ConvolutionDepthWise2D]:
            return False
        # weights shape for a DW convolution is [ch_mult, in_ch, fH, fW]
        ch_mult = layer.getWeights().shape[0]
        qualified = ch_mult > 1

        return qualified

    def isPointwiseConv(layer):
        """
            Returns True/False if the given layer is/is not a 1x1 Convolution Layer
        """
        from Controllers.Parsers.Parser.Convolution2D import Convolution2D
        qualified = True
        qualified = qualified and type(layer) in [Convolution2D]
        qualified = qualified and layer.getKernelSize() == (1,1)
        qualified = qualified and \
                     (layer.padStyle == PadStyle.caffe and layer.getPadding() == (0,0,0,0) or \
                      layer.padStyle == PadStyle.tfvalid)

        return qualified

    def WeightCompensationSeparableConv(dwconv_layer, pointwiseconv_layer):
        # Convolution2D has weights in shape (outCH, inCH, fH, fW)
        # ConvolutionDepthwise2D has weights in shape (inCH, chMultiplier, fH, fW)

        dw_weights = dwconv_layer.getWeights().data
        pw_weights = pointwiseconv_layer.getWeights().data

        ch = dw_weights.shape[0]
        ch_mult = dw_weights.shape[1]
        dw_weights_extend_shape = (ch, ch*ch_mult,
                            dw_weights.shape[2], dw_weights.shape[3])
        dw_weights_extend = np.zeros(dw_weights_extend_shape)

        for c in range(ch):
            pad_l = c*ch_mult
            pad_r = (ch-c-1)*ch_mult
            npad = ((pad_l,pad_r), (0,0), (0,0))
            x = np.pad(dw_weights[c,:,:,:], npad, mode='constant', constant_values=(0, 0))
            dw_weights_extend[c,:,:,:] = x
        # transpose dw_weights_extend to get to shape (outCH, inCH, fH, fW)
        dw_weights_extend = np.transpose(dw_weights_extend, (1, 0, 2, 3))

        final_weights = np.matmul(dw_weights_extend.T, pw_weights.T[0][0]).T

        pointwiseconv_layer.setWeights(final_weights)
        pointwiseconv_layer.loadKernelSize(*dwconv_layer.getKernelSize())
        pointwiseconv_layer.loadStrideSize(*dwconv_layer.getStrideSize())
        pointwiseconv_layer.loadPadding(*dwconv_layer.getPadding())
        pointwiseconv_layer.padStyle = dwconv_layer.padStyle

        return pointwiseconv_layer

    check_again = True
    while check_again:
        g, check_again = fuse_nodes(g, isPointwiseConv, isDepthwiseConv, WeightCompensationSeparableConv, keep_node_a=False)

    return g


def fuseBiasAdd(g):
    from Controllers.Parsers.Parser.InnerProduct import InnerProduct
    from Controllers.Optimizer import fuse_nodes

    """
        Iterates over the graph removing any qualifying fusions for
        bias and add until we are complete.
    """

    def isBiasOrAdd(layer):
        """
            Returns True/False if the given layer is/is not a Bias or Add Layer
        """
        from Controllers.Parsers.Parser.Bias import Bias
        from Controllers.Parsers.Parser.Eltwise import Eltwise
        return (type(layer) == Bias) or \
                ((type(layer) == Eltwise) and (layer.getType() == Eltwise.Type.WSUM))

    def isConvOrFC(layer):
        """
            Returns True/False if the given layer is/is not a Convolution or InnerProduct Layer
        """
        from Controllers.Parsers.Parser.Convolution2D import Convolution2D, ConvolutionDepthWise2D
        return type(layer) in [Convolution2D, ConvolutionDepthWise2D, InnerProduct]

    def PutBiasInConv(layer, absorbed_layer):
        """
            To account for the removal of Bias and Add, we insert the bias
        """
        from Controllers.Parsers.Parser.Bias import Bias
        from Controllers.Parsers.Parser.Eltwise import Eltwise

        if (type(absorbed_layer) == Bias):
            b = absorbed_layer.getBias()
        else:
            # TODO: pull the data out of the Add input
            b = 0
            assert("TensorFlow: Add after Matmul is not supported")

        # Change Bias
        if layer.biasEnabled():
            if b is not None:
                layer.setBias(layer.getBias().data + b)
        else:
            # If there is no bias, it is possible that we will need to now have one
            if b is not None:
                layer.setBiasEnabled(True)
                layer.setBias(np.array(b).astype(np.float16))
        return layer

    check_again = True
    while check_again:
        g, check_again = fuse_nodes(g, isBiasOrAdd, isConvOrFC, PutBiasInConv)

    return g
