# Copyright 2018 Intel Corporation.
# The source code, information and material ("Material") contained herein is
# owned by Intel Corporation or its suppliers or licensors, and title to such
# Material remains with Intel Corporation or its suppliers or licensors.
# The Material contains proprietary information of Intel or its suppliers and
# licensors. The Material is protected by worldwide copyright laws and treaty
# provisions.
# No part of the Material may be used, copied, reproduced, modified, published,
# uploaded, posted, transmitted, distributed or disclosed in any way without
# Intel's prior express written permission. No license under any patent,
# copyright or other intellectual property rights in the Material is granted to
# or conferred upon you, either expressly, by implication, inducement, estoppel
# or otherwise.
# Any license under such intellectual property rights must be express and
# approved by Intel in writing.
import Controllers.EnumController as ec

def getBatchedTensorChannels(numpy_shape):
    assert(len(numpy_shape) == 4)
    return numpy_shape[3]

def getBatchedTensorHeight(numpy_shape):
    assert(len(numpy_shape) == 4)
    return numpy_shape[1]

def getBatchedTensorWidth(numpy_shape):
    assert(len(numpy_shape) == 4)
    return numpy_shape[2]


def fillTensorDimensions(numpy_shape, convertToNCHW=True):
    # Fill out Tensor to hold four dimensions
    from tensorflow import TensorShape

    if isinstance(numpy_shape, TensorShape):
        data = numpy_shape.as_list()
    else:
        data = list(numpy_shape)

    for x in range(4 - len(data)):
        data.insert(0, 1)

    # Convert all tensors to NCHW
    if convertToNCHW:
        data = [data[0], data[3], data[1], data[2]]

    return tuple(data)

def loadTensorSizes(parsedLayers, parsedModelObj):
    tensorDict = {}

    for operation in parsedModelObj.get_operations():
        for x in operation.inputs:
            tensorDict[x.name] = fillTensorDimensions(x.get_shape())

        for x in operation.outputs:
            tensorDict[x.name] = fillTensorDimensions(x.get_shape())

    for layer in parsedLayers:
        layer.loadTensorSizes(tensorDict)


def getTensorShape(tensor_name, model):
    try:
        inputTensor = model.get_tensor_by_name(tensor_name)
    except:
        ec.throw_error(ec.ErrorTable.NoOutputNode, tensor_name)

    shape = inputTensor.get_shape()
    shape_list = shape.as_list()
    # Tensorflow can have None in the batch size field of the
    # input shape, if that is the case then set it to 1
    if None == shape_list[0]:
        shape_list[0] = 1
        shape = shape_list
        inputTensor.set_shape(shape)
    elif None in shape:
        ec.throw_error(ec.ErrorTable.NoOutputNode, "No shape information for output tensor")
    return shape


def getPadding(in_dim, kernel_dim, strides, padding_type):
    def same_padding(in_dim, kernel_dim, strides):
        import numpy as np
        """
        Calculates the output dimension and also the padding required for that dimension.
        :param in_dim: [Height, Width] of Input
        :param kernel_dim: [Height, Width] of Kernel
        :param strides: [Vertical, Horizontal] Strides
        """
        in_dim = np.array(in_dim)
        in_dim[in_dim == None] = 1

        kernel_dim = np.array(kernel_dim)
        strides = np.array(strides)

        out_dim = np.ceil(np.float_(in_dim) / np.float_(strides))
        out_dim = out_dim.astype(int)
        pad_along_y = (out_dim[0]  - 1) * strides[0] + kernel_dim[0] - in_dim[0]
        pad_along_x = (out_dim[1] - 1) * strides[1] + kernel_dim[1] - in_dim[1]
        pad_left = pad_along_x // 2
        pad_right = pad_along_x - pad_left
        pad_top = pad_along_y // 2
        pad_bottom = pad_along_y - pad_top

        return (pad_top, pad_bottom, pad_left, pad_right)

    def valid_padding(in_dim, kernel_dim, strides):
        pad_top = pad_bottom = 0
        pad_left = pad_right = 0

        return (pad_top, pad_bottom, pad_left, pad_right)

    if padding_type == b'VALID':
        return valid_padding(in_dim, kernel_dim, strides)
    elif padding_type == b'SAME':
        return same_padding(in_dim, kernel_dim, strides)
    else:
        return None

def stripTensorName(tensorName):
    if tensorName[-2:] == ':0':
        return tensorName[0:-2]
    else:
        return tensorName

# find tensor from given name and graph operations
def findTensor(tensorName, operations):
    for o in operations:
        for to in o.outputs:
            if to.name == tensorName:
                return to
    return None

# find tensor value
def findTensorValue(tensorName, operations):
    tensor = findTensor(tensorName, operations)
    try:
        values = tensor.eval()
    except:
        values = None

    return values

def getInputNames(op):
    return [x.name for x in op.inputs]

def getOutputNames(op):
    return [x.name for x in op.outputs]
