# Copyright 2018 Intel Corporation.
# The source code, information and material ("Material") contained herein is
# owned by Intel Corporation or its suppliers or licensors, and title to such
# Material remains with Intel Corporation or its suppliers or licensors.
# The Material contains proprietary information of Intel or its suppliers and
# licensors. The Material is protected by worldwide copyright laws and treaty
# provisions.
# No part of the Material may be used, copied, reproduced, modified, published,
# uploaded, posted, transmitted, distributed or disclosed in any way without
# Intel's prior express written permission. No license under any patent,
# copyright or other intellectual property rights in the Material is granted to
# or conferred upon you, either expressly, by implication, inducement, estoppel
# or otherwise.
# Any license under such intellectual property rights must be express and
# approved by Intel in writing.

import itertools
from collections import namedtuple
from math import gcd

import networkx as net_x
import numpy as np

from Controllers.Parsers.Parser.Layer import OriginalName, MangledName
from Controllers.GraphUtils import buildGraph, buildLayerLists
from Controllers.Tensor import Tensor, UnpopulatedTensor
from Controllers.TensorFormat import LayerIOFormat, LayerIOFormatPair, TensorShape, TensorFormat
import Models.Layouts as Layouts
from Controllers.Parsers.Parser.Conversion import Conversion

IdxIOFormat = namedtuple("IdxIOFormat", "idx io_format")
CombScore = namedtuple("CombScore", "comb score")
TFormatScore = namedtuple("TFormatScore", "t_format score")

def get_enclosure_request_lcm(enclosure_requests):
    """
        The lowest_common_enclosure request is commputed as follows, in accordance
        with the definition of the EnclosureRequest (see TensorFormat.EnclosureRequest):
        1. For the N, H, W axes the lcm_enclosure is the lowest common multiple
           of the requested enclosures.
        2. For the C axis the lcm_enclosure is the maximum of the requeted enclosures.

        return: lcm_enclosure
    """
    lcm_enclosure = enclosure_requests[0]
    for enclosure_req in enclosure_requests:
        lcm_enclosure = TensorShape(*[
            max(lcm_req_axis, enc_req_axis) if axis == Layouts.Axes.C else
            lcm_req_axis * enc_req_axis // (gcd(lcm_req_axis, enc_req_axis))
            for axis,(lcm_req_axis, enc_req_axis) in enumerate(zip(lcm_enclosure, enclosure_req))
            ])

    return lcm_enclosure

def score_combination(combination, tensor_shape):
    lcm_formats = get_lcm_tensor_formats(combination, tensor_shape)

    num_conversions_req = 0
    producer_format_chosen = False
    producer_fm = combination[0]
    for t_layout, t_formats in lcm_formats.items():
        if producer_fm.io_format.layout != t_layout:
            num_conversions_req = num_conversions_req + len(t_formats)
        else:
            for t_format_score in t_formats:
                if (producer_fm.io_format.can_produce(t_format_score.t_format)
                        and not producer_format_chosen):
                    producer_format_chosen = True
                else:
                    num_conversions_req = num_conversions_req + t_format_score.score

    return num_conversions_req

def get_lcm_tensor_formats(io_formats, tensor_shape):
    layouts = set([idx_format.io_format.layout for idx_format in io_formats])
    lcm_formats = {layout: [] for layout in layouts}

    for layout in layouts:
        formats_unsatisfied = [idx_format.io_format for idx_format in io_formats
            if idx_format.io_format.layout == layout]

        num_unsatisfied = len(formats_unsatisfied)
        combination_size = num_unsatisfied
        while num_unsatisfied:
            format_combs = list(itertools.combinations(formats_unsatisfied, combination_size))

            enclosure_lcms = [
                get_enclosure_request_lcm(
                    [io_format.enclosure_request for io_format in format_comb])
                for format_comb in format_combs
                ]

            lcm_tensor_formats = [
                TensorFormat(layout, tensor_shape,
                    TensorFormat.compute_enclosure_shape(tensor_shape, enclosure_req_lcm))
                for enclosure_req_lcm in enclosure_lcms
                ]

            # Eliminate tensor formats that do NOT satisfy the combination.
            pruned_combinations = [(f_comb, t_format) for (f_comb, t_format) in
                zip(format_combs, lcm_tensor_formats) if all(
                    [io_format.can_consume(t_format) for io_format in f_comb])
                ]

            if len(pruned_combinations) != 0:
                # Eliminate formats that are satisfied by the "best" of the current
                # lcm_formats.
                chosen_comb = pruned_combinations[0]
                formats_unsatisfied = [io_format for io_format in formats_unsatisfied
                    if not io_format.can_consume(chosen_comb[1])]

                score = num_unsatisfied - len(formats_unsatisfied)
                lcm_formats[layout].append(TFormatScore(chosen_comb[1], score))
                num_unsatisfied = len(formats_unsatisfied)
                combination_size = num_unsatisfied
            else:
                combination_size = combination_size - 1

    return lcm_formats

def generate_format_combinations(producer_formats, consumers_formats):
    format_lists = [producer_formats]
    for consumer_formats in consumers_formats:
        if (not any([consumer_formats == fmt for fmt in format_lists])):
            format_lists.append(consumer_formats)

    combinations = [[fmt] for fmt in format_lists[0]]

    if len(format_lists) > 0:
        for io_formats in format_lists[1:]:
            combinations = [comb + [fmt] for comb in combinations for fmt in io_formats]

    return combinations

def select_tensors_format(parsedLayers):
    graph = buildGraph(parsedLayers, contract_blobs=False)

    for current_node_name in net_x.lexicographical_topological_sort(graph):
        current_node = graph.node[current_node_name]
        if current_node['type'] == 'BLOB':
            current_tensor = current_node['ref']
            current_tensor_shape = current_tensor.shape

            prod_layers = [graph.node[layer_name]['ref'] for layer_name
                            in graph.predecessors(current_node_name)]
            assert(len(prod_layers) == 1)
            prod_layer = prod_layers[0]
            cons_layers = [graph.node[layer_name]['ref'] for layer_name
                            in graph.successors(current_node_name)]

            prod_formats = [
                IdxIOFormat(format_id, io_format[current_node_name])
                for (format_id, io_format) in enumerate(prod_layer.getSelectedIOFormats())
                ]

            cons_formats = [
                [IdxIOFormat(format_id, io_format[current_node_name])
                for (format_id, io_format) in enumerate(consumer.getSelectedIOFormats())]
                for consumer in cons_layers
                ]

            format_combinations = generate_format_combinations(prod_formats,
                                                               cons_formats)
            best_combination = None
            for comb in format_combinations:
                score = score_combination(comb, current_tensor_shape)

                if best_combination is None:
                    best_combination = CombScore(comb, score)
                else:
                    if best_combination.score > score:
                        best_combination = CombScore(comb, score)
                if score == 0:
                    break

            assert(best_combination is not None)

            lcm_formats = get_lcm_tensor_formats(best_combination.comb,
                                                          current_tensor_shape)

            # Set the producer IO formats.
            producer_best_format = best_combination.comb[0]

            prod_layer.setSelectedIOFormats(
                [prod_layer.getSelectedIOFormats()[producer_best_format.idx]])

            # Holds the tensor formats for the tesors that have to be produced in order
            # to satisfy all consumers. The first tensor for in the list is the
            # format that is going to be produced by the producer.
            final_tensor_formats = [
                t_format_score for format_scores in lcm_formats.values()
                               for t_format_score in format_scores
                ]
            final_tensor_formats = sorted(final_tensor_formats, key=lambda fs: fs.score)

            tensors_to_produces = [
                [t_format, tensor] for t_format, tensor in \
                    zip(final_tensor_formats, [None,] * len(final_tensor_formats))
                ]

            # Set the produced tensor according to the best format (i.e. the format
            # that satisfies the producer and most consumers).
            best_tensor_format = None
            for tensor in tensors_to_produces:
                if producer_best_format.io_format.can_produce(tensor[0].t_format):
                    best_tensor_format = tensor[0].t_format
                    tensor[1] = current_tensor

            current_tensor.setLayout(best_tensor_format.layout)
            if not best_tensor_format.is_compact():
                enclosure_shape = best_tensor_format.get_enclosure_shape()
                enclosure = UnpopulatedTensor(shape=enclosure_shape)
                enclosure.setName(MangledName(OriginalName("satisfier")))
                enclosure.setLayout(best_tensor_format.layout)
                enclosure.setDatatype(np.float16)
                current_tensor.place(enclosure, Tensor.ORIGIN)

            for consumer_name in list(graph.successors(current_node_name)):
                consumer_layer = graph.node[consumer_name]['ref']

                consumer_solution_tensor = [
                    (idx, tensor_node) for idx, tensor_node in enumerate(tensors_to_produces)
                    if any([io_format[current_node_name].can_consume(tensor_node[0].t_format)
                            for io_format in consumer_layer.getSelectedIOFormats()])
                    ]

                consumer_solution_tensor = sorted(consumer_solution_tensor,
                                                    key=lambda fs_node: fs_node[1][0].score)

                consumer_solution_tensor = consumer_solution_tensor[0]
                consumer_solution_tensor_idx = consumer_solution_tensor[0]
                solution_t_format = consumer_solution_tensor[1][0].t_format

                can_consume = [io_format for io_format in consumer_layer.getSelectedIOFormats()
                               if io_format[current_node_name].can_consume(solution_t_format)]
                consumer_layer.setSelectedIOFormats(can_consume)

                solution_node = consumer_solution_tensor[1][1]
                if solution_node is None:
                    converted_tensor_name = '{}_converted'.format(
                        current_tensor.getName().stringifyOriginalName())
                    converted_tensor = UnpopulatedTensor(shape=solution_t_format.shape)
                    converted_tensor.setName(MangledName(OriginalName(converted_tensor_name)))
                    converted_tensor.setLayout(solution_t_format.layout)
                    graph.add_node(converted_tensor.getName().stringifyName(),
                                    type="BLOB", ref=converted_tensor)
                    # Set the node for the produced tensor.
                    tensors_to_produces[consumer_solution_tensor_idx][1] = converted_tensor
                    solution_node = tensors_to_produces[consumer_solution_tensor_idx][1]
                    graph.add_edge(solution_node.getName().stringifyName(), consumer_name)
                    graph.remove_edge(current_node_name, consumer_name)

                    consumer_layer.inputTensorNames = [
                        t_name if t_name.stringifyName() != current_node_name
                               else converted_tensor.getName() for
                        t_name in consumer_layer.inputTensorNames]

                    consumer_layer.setInputTensors([
                        tensor if tensor.getName().stringifyName() != current_node_name
                               else converted_tensor
                        for tensor in consumer_layer.getInputTensors()])

                    consumer_layer.loadInputTensorSizes([
                        tensor.getShape() for tensor in consumer_layer.getInputTensors()])

                    # Place conveted tensor in enclosure if needed.
                    if not solution_t_format.is_compact():
                        enclosure_shape = solution_t_format.get_enclosure_shape()
                        enclosure = UnpopulatedTensor(shape=enclosure_shape)
                        enclosure.setName(MangledName(OriginalName("satisfier")))
                        enclosure.setLayout(solution_t_format.layout)
                        enclosure.setDatatype(np.float16)
                        converted_tensor.place(enclosure, Tensor.ORIGIN)

                    conversion_layer_name = "{}_convert".format(
                        prod_layer.getName().stringifyOriginalName())
                    conversion_layer = Conversion(conversion_layer_name,
                                    [current_tensor.getName()],
                                    [converted_tensor.getName()])
                    conversion_layer.setInputTensors([current_tensor])
                    conversion_layer.setOutputTensors([converted_tensor])
                    conversion_layer.loadInputTensorSizes([current_tensor.getShape()])
                    conversion_layer.loadOutputTensorSizes([converted_tensor.getShape()])
                    graph.add_node(conversion_layer.getStringifiedName(),
                                type="OP", ref=conversion_layer)

                    graph.add_edge(conversion_layer.getStringifiedName(),
                               converted_tensor.getName().stringifyName())

                    graph.add_edge(current_node_name, conversion_layer.getStringifiedName())

                elif solution_node.getName().stringifyName() != current_node_name:
                    graph.add_edge(solution_node.getName().stringifyName(), consumer_name)

                    consumer_layer.inputTensorNames = [
                        t_name if t_name.stringifyName() != current_node_name
                               else solution_node.getName() for
                        t_name in consumer_layer.inputTensorNames]

                    consumer_layer.setInputTensors([
                        tensor if tensor.getName().stringifyName() != current_node_name
                               else solution_node
                        for tensor in consumer_layer.getInputTensors()])

                    consumer_layer.loadInputTensorSizes([
                        tensor.getShape() for tensor in consumer_layer.getInputTensors()])

    return buildLayerLists(graph)
