# Copyright 2018 Intel Corporation.
# The source code, information and material ("Material") contained herein is
# owned by Intel Corporation or its suppliers or licensors, and title to such
# Material remains with Intel Corporation or its suppliers or licensors.
# The Material contains proprietary information of Intel or its suppliers and
# licensors. The Material is protected by worldwide copyright laws and treaty
# provisions.
# No part of the Material may be used, copied, reproduced, modified, published,
# uploaded, posted, transmitted, distributed or disclosed in any way without
# Intel's prior express written permission. No license under any patent,
# copyright or other intellectual property rights in the Material is granted to
# or conferred upon you, either expressly, by implication, inducement, estoppel
# or otherwise.
# Any license under such intellectual property rights must be express and
# approved by Intel in writing.


import numpy as np
from ctypes import *
import math


from Models.HwDescriptor import HwDescriptor, HwDescriptorList
from Controllers.Soh import heightSolution, heightSolutionWithPooling
from Controllers.Parsers.Parser.Hw import *

def round_up(x, mult):
    return ((x + mult -1) // mult) * mult

def getPadding(inputSize, outputSize, kernelDimX, kernelDimY, kernelStride):
    valid_out_x = math.ceil((inputSize[3] - kernelDimX + 1) / kernelStride)
    valid_out_y = math.ceil((inputSize[2] - kernelDimY + 1) / kernelStride)

    pad_along_x = (outputSize[3] - 1) * kernelStride + kernelDimX - inputSize[3]
    pad_along_y = (outputSize[2] - 1) * kernelStride + kernelDimY - inputSize[2]
    pad_left = pad_along_x // 2
    pad_right = pad_along_x - pad_left
    pad_top = pad_along_y // 2
    pad_bottom = pad_along_y - pad_top
    return (outputSize[3] != valid_out_x or outputSize[2] != valid_out_y, pad_left, pad_right, pad_top, pad_bottom)

def getStrides(inputSize, outputSize, bytesPerPixel):
    inputStrideWidth = bytesPerPixel
    inputStrideHeight = ((inputSize[3] * inputStrideWidth + 15) // 16) * 16
    inputStrideZ = inputStrideHeight * inputSize[2]
    outputStrideWidth = bytesPerPixel
    outputStrideHeight = ((outputSize[3] * outputStrideWidth + 15) // 16) * 16
    outputStrideZ = outputStrideHeight * outputSize[2]

    return (inputStrideWidth, inputStrideHeight, inputStrideZ), (outputStrideWidth, outputStrideHeight, outputStrideZ)

def getStridesAndLayout(stage,inputTensor,outputTensor,splits,bytesPerPixel):
    """Returnst inputStride and outputStride tuple
       The strides strictly come from the enclosures of the tensors. The strides are logical, in the sense that they do
       no represent memory layouts
    """

    #TODO: re-use less code .... Put this in a better place....

    #TODO: CHECK is padding over the WIDTH dimension requested as an enclosure or encoded in the tensor dimension?
    #TODO: Do we need to check the Width Height from the enclosure too? Logically yes, but why was it not done until now?

    _, inputChannelLayout,inputHeightLayout,inputWidhtLayout  = inputTensor.getTopEncloserRecursive().getShape()
    _, outputChannelLayout, outputHeightLayout, outputWidthLayout = outputTensor.getTopEncloserRecursive().getShape()

    inputChannelLayout = inputChannelLayout

    inputStrideWidth = bytesPerPixel
    outputStrideWidht = bytesPerPixel

    inputStrideHeight = round_up(inputWidhtLayout * inputStrideWidth, 16)
    outputStrideHeight = round_up(outputWidthLayout * outputStrideWidht, 16)

    inputStrideChannels = inputStrideHeight * inputHeightLayout
    outputStrideChannels = outputStrideHeight * outputHeightLayout

    inputStrideBatch = inputStrideChannels * inputChannelLayout
    outputStrideBatch = outputStrideChannels * outputChannelLayout



    inputStrides = (inputStrideBatch,inputStrideChannels,inputStrideHeight,inputStrideWidth)
    outputStrides = (outputStrideBatch,outputStrideChannels,outputStrideHeight,outputStrideWidht)

    inputLayout = (1, inputChannelLayout, inputHeightLayout, inputWidhtLayout)
    outputLayout = (1, outputChannelLayout, inputHeightLayout, inputWidhtLayout)

    return inputLayout,outputLayout,inputStrides,outputStrides



def trueRowAdaptationB0(maxOutputLines, pad_left, pad_right, pad_top, pad_bottom, inputSize, kernelHeight):

    if pad_bottom:
        max_size = 2048
    elif (pad_left or pad_right):
        max_size = 4096
    else:
        max_size = round_up(inputSize[3], 16) * (maxOutputLines-1 + kernelHeight-1)
    maxOutputLines = max_size//round_up(inputSize[3], 16) -kernelHeight + 2

    return maxOutputLines

def compile_conv(stage):

    bytesPerPixel = 2# np.dtype(stage.getWeights().dtype).itemsize

    # Basic checks
    assert len(stage.inputTensors) == 1
    assert len(stage.outputTensors) == 1

    # Get solution and splits
    solution = stage.getSolution()
    splits = stage.getSplitOverC()
    # solution[2].sort(key = lambda tup: tup[1])
    scheduledInputDimZ, scheduledOutputDimZ, tiles = solution

    # Get input and output tensors
    inputTensor = stage.inputTensors[0]
    outputTensor = stage.outputTensors[0]
    inputSize = inputTensor.getShape()
    outputSize = outputTensor.getShape()

    #assert for batch=1
    assert(inputSize[0] == 1)
    assert(outputSize[0] == 1)

    inputLayout,outputLayout,inputStride, outputStride = getStridesAndLayout(stage,inputTensor,outputTensor,splits,bytesPerPixel)

    # Calculate the strides of the InOut buffers
    completeHwDescList = []

    # Default for relu
    reluOnShaveAccumulation = False
    reluNegSlope = 0
    reluPosSlope = 1

    reluXOnShaveAccumulation = False
    XofReluX = 0

    padEnable, pad_left, pad_right, pad_top, pad_bottom = getPadding(inputSize, outputSize, stage.kernelWidth, stage.kernelHeight, stage.strideWidth)
    # Cycle input channel splits
    for sodIndex in range(splits):
        #Calculate in/out strides
        #inputStride, outputStride = getStrides(inputSize, outputSize, bytesPerPixel)

        # Calculate buffers Size . 
        #TODO: currently am asuming the batches are continous. basically, the size, is a channelStride*batch
        inputBufferSize = inputStride[0] * inputSize[0]
        outputBufferSize = outputStride[0] * outputSize[0]
        # tapsBufferSize = scheduledInputDimZ * scheduledOutputDimZ * stage.kernelWidth * stage.kernelHeight * bytesPerPixel
        # biasBufferSize = scheduledOutputDimZ * bytesPerPixel if stage.biasEnabled() is not None else 0

        heightSplits = []
        # Calculate the maximum number of output lines
        if stage.isStreaming()  and stage.isSplitOverH():
            maxOutputChannelsInDescr = max(tiles, key=lambda x: x[0])[0]
            bytesPerFullDepthSlice = bytesPerPixel * maxOutputChannelsInDescr * round_up(outputSize[3], 8)
            maxOutputLines = stage.getCMXforStreaming() // bytesPerFullDepthSlice
        else:
            maxOutputLines = outputSize[2]

        # Get the number of split over height
        if isinstance(stage, HwConvolutionPooling):
            # For conv3x3s1p1 and fused 2x2s2 pooling, we need 4 extra lines
            # Also, for this specific case, the maxOutputLines is doubled, because the output
            # is reduced by a factor of 2
            assert(stage.pad_top == stage.pad_bottom)
            heightSplits = heightSolutionWithPooling(inputSize[2], stage.kernelHeight, stage.strideHeight, stage.pad_top, maxOutputLines)
        else:
            # For convolution without fused pooling
            # The following is not correct for convolution. We cannot have selective zero padding
            # pad = (stage.kernelHeight // 2 if pad_top > 0 else 0, stage.kernelHeight // 2 if pad_bottom > 0 else 0)
            assert(stage.pad_top == stage.pad_bottom)
            assert(stage.pad_top == 0 or stage.pad_top == (stage.kernelHeight // 2))
            pad = (stage.pad_top, stage.pad_bottom)
            heightSplits = heightSolution(inputSize[2], stage.kernelHeight, stage.strideHeight, pad, maxOutputLines)


        sohGroups = []
        for sohIndex, heightSplitSol in enumerate(heightSplits):

            trueInputNeeded, _, junkBefore, junkAfter, inputStartIndex, _, outputStartIndex, _ = heightSplitSol
            outChanOffset = 0

            sohGroup = HwDescriptorList()

            for tileIndex, (outChans, mode) in enumerate(tiles):
                if tileIndex == (len(tiles) - 1) and \
                    sohIndex == len(heightSplits) - 1 and \
                    	sodIndex == splits - 1:
                    	lastTile = True
                else:
                    lastTile = False

                conv = HwDescriptor(0, 0, 1, 0, mode, tileIndex, lastTile, stage.getName())
                conv.topOutputJunk = junkBefore
                conv.bottomOutputJunk = junkAfter

                if round_up(inputSize[3], 8) * inputSize[2] * (scheduledInputDimZ) * bytesPerPixel < 128*1024:
                    if tileIndex == 0:
                        conv.reuseData = stage.rud
                    else:
                        if tiles[tileIndex-1][1] != mode:
                            conv.reuseData = 0
                        else:
                            conv.reuseData = 1

                # Input in interleaved
                if inputTensor.getLayout() == (0, 2, 1, 3):
                    conv.setupInterleaved(True)
                    conv.setupInterleavedInput(True)

                    split_by_input_sz = sodIndex * scheduledInputDimZ * inputStride[2]
                    split_by_height_sz = (scheduledInputDimZ * splits) * inputStride[2] * inputStartIndex
                    offset = split_by_input_sz + split_by_height_sz

                    conv.setupInput(offset, inputSize[3], trueInputNeeded, scheduledInputDimZ, inputStride[2] * inputLayout[1], inputStride[2])

                else:
                    #A in interleaved a channel i a 'line'
                    split_by_input_sz = sodIndex * scheduledInputDimZ * inputStride[1]
                    split_by_height_sz = inputStride[2] * inputStartIndex
                    offset = split_by_input_sz + split_by_height_sz
                    conv.setupInput(offset, inputSize[3], trueInputNeeded, scheduledInputDimZ, inputStride[2], inputStride[1])

                # if stage.isConcat():
                #     totalOutChans = stage.totalOutChans
                # else:
                #     totalOutChans = sum([outChans for outChans, _ in tiles])

                # if newOutputDimZ > totalOutChans and newOutputDimZ == outputTensor.getTopEncloserRecursive().shape[1]:
                #     totalOutChans = max(newOutputDimZ, totalOutChans)

                #This is a bit tricky. Since we may need to output with a larger stride. Since CMX is precious, we should NEVER EVER 
                #Output with padding of stride inside CMX. In CMX we should output compact, and only the DMA's should output strided.
                #In Row-Major, we will output width*actualChannels as LineStride , and the DMA should take over as width*stridedChannels
                #in planar , we don't really care (TODO: check if really) , since the planes are one after the other, and we don't care what's there
                #TODO:: output compact also in widht/height :(
                #Upper note is both for input and output
                
                # output in interleaved
                if outputTensor.getLayout() == (0, 2, 1, 3):
                    conv.setupInterleavedOutput(True)

                    if stage.isStreaming() :
                        # if we are streaming, we will write to CMX in compacty mode. The DMA will handle the strides
                        outputStrideOnZ =  outputStride[2] * scheduledOutputDimZ
                    else:
                        # if we ain't streaming, we need to write to CMX as the layout dictates
                        outputStrideOnZ =  outputStride[2] * outputLayout[1]

                    offset = outputStartIndex * outputLayout[1] * outputStride[2] + outChanOffset * outputStride[2]

                    conv.setupOutput(offset, outputSize[3], outputSize[2], outChans,outputStrideOnZ, outputStride[2])
                else:
                    offset = outputStartIndex * outputStride[2] + outChanOffset * outputStride[1]
                    conv.setupOutput(offset, outputSize[3], outputSize[2], outChans, outputStride[2], outputStride[1])

                taps_processed = sum([a for a, _ in tiles[:tileIndex]])
                # taps_total = sum([a for a, _ in tiles])
                taps_total = round_up(scheduledOutputDimZ,8)

                split_by_input_taps_sz = bytesPerPixel * taps_total * scheduledInputDimZ * stage.kernelWidth * stage.kernelHeight
                conv.setupConvolutionCoefficients(0,
                        sodIndex * split_by_input_taps_sz +
                        bytesPerPixel * taps_processed * scheduledInputDimZ * stage.kernelWidth * stage.kernelHeight,
                        stage.kernelWidth, stage.kernelHeight, stage.strideWidth)
                if padEnable:
                    conv.setupPadding(0)

                # if sohIndex == 0 and (stage.bias is not None):
                if sodIndex == 0 and stage.biasEnabled():
                    conv.setupBias(outChanOffset * bytesPerPixel)

                if isinstance(stage, HwConvolutionPooling) and (sodIndex == 0):
                    # If we have fused pooling and more that one depth groups, then
                    # we need not to fuse the pooling
                    if stage.type == Pooling.Type.MAX:
                        poolType = 0
                    else:
                        raise ValueError('Average pooling fusion with conv not supported')

                    conv.setupConvolutionPooling(poolType, stage.pooling_kernelWidth, stage.pooling_kernelHeight)

                # Since we run this in hardware, we need to convert postops (and whatever else) that are
                # supported by the hardware
                postOp = [x for x in stage.getPostOp() if isinstance(x, ReLU)]
                if len(postOp) > 1:
                    raise ValueError("Only one post operation is supported. Found {}".format(postOp))
                elif len(postOp) == 1:
                    # General form of hardware Parametrized ReLU (PRelu):
                    #                     / a0*x, if x < t0 ('negative' slope)
                    # f(x; t0, a0, a1) = |
                    #                     \ a1*x, if x >= t0 ('positive' slope)


                    if splits > 1:
                        if hasattr(stage.post_op[0], 'reluX'):
                            XofReluX = stage.post_op[0].reluX
                            if (XofReluX != 0):
                                reluXOnShaveAccumulation = True
                                conv.setupReluX(stage.post_op[0].reluX)

                        if (XofReluX == 0):
                            if hasattr(stage.post_op[0], 'negativeSlope'):
                                reluOnShaveAccumulation = True
                                reluNegSlope = stage.post_op[0].negativeSlope

                    else:
                        reluOnShaveAccumulation = False
                        reluXOnShaveAccumulation = False

                        if hasattr(stage.post_op[0], 'reluX'):
                            XofReluX = stage.post_op[0].reluX
                            if (XofReluX != 0):
                                conv.setupReluX(stage.post_op[0].reluX)

                        if (XofReluX == 0):
                            if hasattr(stage.post_op[0], 'negativeSlope'):
                                reluNegSlope = stage.post_op[0].negativeSlope
                                # Check that the scaling will be 10 bits signed integer
                                if (reluNegSlope == 0):
                                    conv.setupRelu(0, 0, 1)
                                else:
                                    assert (abs(int(round(1 / reluNegSlope))) < 2 ** 9)
                                    stage.setReLUScale(reluNegSlope)
                                    conv.setupRelu(0, 1, int(round(1 / reluNegSlope)))
                                    conv.setupScale(outChanOffset * bytesPerPixel)
                            else:
                                conv.setupRelu(0, 0, 1)

                outChanOffset += outChans
                conv.sohGroup = sohIndex
                conv.sodGroup = sodIndex
                sohGroup.pushDescriptor(conv)

            sohGroups.append(sohGroup)
        completeHwDescList.append(sohGroups)

    # Transpose the descriptors so that we have a list of SoH,
    # where each element will be a SoD sublist.
    transposedDescrList = list(map(list, zip(*completeHwDescList)))

    # Flatten the descriptors
    finalHwDescList = HwDescriptorList()

    for soh in transposedDescrList:
        for sod in soh:
            for item in sod.descList:
                finalHwDescList.pushDescriptor(item)

    # Convert taps to hardware
    weights = stage.getWeights()
    taps = taps_hwck_to_cnnhw(np.transpose(weights.data, (2,3,1,0)), scheduledInputDimZ * splits, scheduledOutputDimZ)
    if splits > 1:
        taps = regroup_taps(taps, splits)

    if stage.biasEnabled():
        bias = stage.getBias().data
    else:
        bias = None

    # TODO: Implement ME!!
    scale = None

    #associate taps to the layer
    weights.data = taps
    weights.shape = (taps.shape[0]*taps.shape[3], scheduledInputDimZ * splits, weights.shape[2], weights.shape[3])
    #TODO: weights size can be different from data size in bytes
    tapsBufferSize = np.prod(taps.shape) * bytesPerPixel
    biasBufferSize = scheduledOutputDimZ * bytesPerPixel

    concatOffset, finalOutputBufferSize = stage.getConcatParameters()

    if finalOutputBufferSize == 0:
        finalOutputBufferSize = outputBufferSize

    #TODO: remove , buffer-style strides for return 
    inputBufStride = ( inputStride[3],inputStride[2],inputStride[1])
    otuputBufStride = ( outputStride[3], outputStride[2], outputStride[1])

    solution = (inputBufStride,
                otuputBufStride,
                (scheduledInputDimZ, scheduledOutputDimZ, concatOffset, reluOnShaveAccumulation, reluNegSlope, reluPosSlope, reluXOnShaveAccumulation, XofReluX),
                (inputBufferSize, finalOutputBufferSize, tapsBufferSize, biasBufferSize),
                finalHwDescList,
                taps,
                bias,
                scale)

    return solution


def compile_pooling(stage):

        bytesPerPixel = 2

        # Basic checks
        assert len(stage.inputTensors) == 1
        assert len(stage.outputTensors) == 1
        assert stage.kernelWidth == stage.kernelHeight
        assert stage.strideWidth == stage.strideHeight

        splits = 1 # which means no splits
        solution = stage.getSolution()
        scheduledInputDimZ, scheduledOutputDimZ, tiles = solution

        # Get input and output tensors
        inputTensor = stage.inputTensors[0]
        outputTensor = stage.outputTensors[0]
        inputSize = inputTensor.getShape()
        outputSize = outputTensor.getShape()

        inputLayout,outputLayout,inputStride, outputStride = getStridesAndLayout(stage,inputTensor,outputTensor,splits,bytesPerPixel)

        completeHwDescList = []

        padEnable, pad_left, pad_right, pad_top, pad_bottom = getPadding(inputSize, outputSize, stage.kernelWidth, stage.kernelHeight, stage.strideWidth)

        for sodIndex in range(splits):

            inputBufferSize = inputStride[0] * inputSize[0]
            outputBufferSize = outputStride[0] * outputSize[0]

            heightSplits = []
            # Calculate the maximum number of output lines
            if stage.isStreaming()  and stage.isSplitOverH():
                maxOutputChannelsInDescr = max(tiles, key=lambda x: x[0])[0]
                bytesPerFullDepthSlice = bytesPerPixel * maxOutputChannelsInDescr * round_up(outputSize[3], 8)
                maxOutputLines = stage.getCMXforStreaming() // bytesPerFullDepthSlice
            else:
                maxOutputLines = outputSize[2]

            if padEnable == 1:
                maxOutputLines = trueRowAdaptationB0(maxOutputLines, pad_left, pad_right, pad_top, pad_bottom, inputSize, stage.kernelHeight)

            pad = (stage.kernelHeight // 2 if pad_top > 0 else 0, stage.kernelHeight // 2 if pad_bottom > 0 else 0)
            heightSplits = heightSolution(inputSize[2], stage.kernelHeight, stage.strideHeight, pad, maxOutputLines)

            # trueInputNeeded, outputWithJunk, junkBefore, junkAfter, inputStartIndex, inputEndIndex, outputStartIndex, outputEndIndex
            # print(heightSplits)

            sohGroups = []
            for sohIndex, heightSplitSol in enumerate(heightSplits):

                trueInputNeeded, _, junkBefore, junkAfter, inputStartIndex, _, outputStartIndex, _ = heightSplitSol
                outChanOffset = 0

                sohGroup = HwDescriptorList()

                for tileIndex, (outChans, mode) in enumerate(tiles):
                    if tileIndex == (len(tiles) - 1) and \
                        sohIndex == len(heightSplits) - 1 and \
                        sodIndex == splits - 1:
                        lastTile = True
                    else:
                        lastTile = False

                    pool = HwDescriptor(0, 0, 1, 0, mode, tileIndex, lastTile, stage.getName())
                    pool.topOutputJunk = junkBefore
                    pool.bottomOutputJunk = junkAfter

                    chansSoFar = sum([outChans for outChans, _ in tiles[0:tileIndex]])
                    # if stage.isConcat():
                    #     totalOutChans = stage.totalOutChans
                    # else:
                    #     totalOutChans = sum([outChans for outChans, _ in tiles])

                    # if newOutputDimZ > totalOutChans and newOutputDimZ == outputTensor.getTopEncloserRecursive().shape[1]:
                    #     totalOutChans = max(newOutputDimZ, totalOutChans)

                    # Input in interleaved
                    if inputTensor.getLayout() == (0, 2, 1, 3):
                        pool.setupInterleaved(True)
                        pool.setupInterleavedInput(True)

                        split_by_input_sz = chansSoFar * inputStride[2]
                        split_by_height_sz = inputLayout[1] * inputStride[2] * inputStartIndex
                        offset = split_by_input_sz + split_by_height_sz
                        pool.setupInput(offset, inputSize[3], trueInputNeeded, outChans, inputStride[2] * inputLayout[1], inputStride[2])
                    else:
                        split_by_input_sz = chansSoFar * inputStride[1]
                        split_by_height_sz = inputStride[2] * inputStartIndex
                        offset = split_by_input_sz + split_by_height_sz
                        pool.setupInput(offset, inputSize[3], trueInputNeeded, outChans, inputStride[2], inputStride[1])

                    # output in interleaved
                    if outputTensor.getLayout() == (0, 2, 1, 3):
                        pool.setupInterleavedOutput(True)
                        if stage.isStreaming() :
                            # if we are streaming, we will write to CMX in compacty mode. The DMA will handle the strides
                            outputStrideOnZ =  outputStride[2] * scheduledOutputDimZ
                        else:
                            # if we ain't streaming, we need to write to CMX as the layout dictates
                            outputStrideOnZ =  outputStride[2] * outputLayout[1]

                        offset = outputStartIndex * outputLayout[1] * outputStride[2] + outChanOffset * outputStride[2]

                        pool.setupOutput(offset, outputSize[3], outputSize[2], outChans, outputStrideOnZ, outputStride[2])
                    else:
                        offset = outputStartIndex * outputStride[2]+ outChanOffset * outputStride[1]
                        pool.setupOutput(offset, outputSize[3], outputSize[2], outChans, outputStride[1], outputStride[2])

                    if stage.type == Pooling.Type.MAX:
                        poolType = 0
                    else:
                        poolType = 1
                    pool.setupPooling(poolType, stage.kernelWidth, stage.kernelHeight, stage.strideWidth)
                    padType = 0

                    if padEnable:
                        # Repeat padding is not used in CNNs at all.
                        if pad_left > 0:
                            padType |= 0x08
                        if pad_right > 0:
                            padType |= 0x01
                        if pad_top > 0:
                            padType |= 0x04
                        if pad_bottom > 0:
                            padType |= 0x02
                        pool.setupPadding(padType)

                    outChanOffset += outChans
                    pool.sohGroup = sohIndex
                    pool.sodGroup = sodIndex
                    sohGroup.pushDescriptor(pool)

                sohGroups.append(sohGroup)
            completeHwDescList.append(sohGroups)

        # Transpose the descriptors so that we have a list of SoH,
        # where each element will be a SoD sublist.
        transposedDescrList = list(map(list, zip(*completeHwDescList)))

        # Flatten the descriptors
        finalHwDescList = HwDescriptorList()

        for soh in transposedDescrList:
            for sod in soh:
                for item in sod.descList:
                    finalHwDescList.pushDescriptor(item)

        concatOffset, finalOutputBufferSize = stage.getConcatParameters()

        if finalOutputBufferSize == 0:
            finalOutputBufferSize = outputBufferSize

        #TODO: remove , buffer-style strides for return 
        inputBufStride = ( inputStride[3],inputStride[2],inputStride[1])
        otuputBufStride = ( outputStride[3], outputStride[2], outputStride[1])
        
        tapsBufferSize = 0
        biasBufferSize = 0 #TODO pool can actually have scale and bias :/
        solution = (inputBufStride,
                    otuputBufStride,
                    (scheduledInputDimZ, scheduledOutputDimZ, concatOffset, False, 0.0, 1.0,False,0.0),
                    (inputBufferSize, finalOutputBufferSize, tapsBufferSize, biasBufferSize),
                    finalHwDescList,
                    None,
                    None,
                    None)

        return solution

def compile_fc(stage):


    bytesPerPixel = 2# np.dtype(stage.getWeights().dtype).itemsize

    # Basic checks
    assert len(stage.inputTensors) == 1
    assert len(stage.outputTensors) == 1

    # Get input and output tensors
    inputTensor = stage.inputTensors[0]
    outputTensor = stage.outputTensors[0]
    inputSize = inputTensor.getShape()
    outputSize = outputTensor.getShape()

    # Get solution and splits
    solution = stage.getSolution()
    splits = stage.getSplitOverC()

    newInputDimZ, newOutputDimZ, tiles = solution

    assert(inputSize[1] % splits == 0)

    completeHwDescList = []

    #Calculate in/out strides
    inputStrideX = 0
    inputStrideY = 0
    inputStrideZ = 16
    outputStrideX = 0
    outputStrideY = 0
    outputStrideZ = 16
    inputBufferSize = newInputDimZ * inputStrideZ
    outputBufferSize = newOutputDimZ * outputStrideZ

    tapsBufferSize = newInputDimZ * newOutputDimZ * bytesPerPixel
    biasBufferSize = 0
    if stage.biasEnabled():
        biasBufferSize = newOutputDimZ * bytesPerPixel
    hwDescList = HwDescriptorList()
    genTileIndex = 0
    outputOffset = 0
    for tileIndex, subTiles in enumerate(tiles):
        lastTile = tileIndex == len(tiles) - 1
        inputOffset = 0

        # Detect how many of the output channels the current subTiles
        # are generating is actual/real. This is needed for the early interrupts
        # workaround.
        outChansSoFar = sum([outChans[0][1] for outChans in tiles[0:tileIndex+1]])
        actualOutputChannels = outputSize[1]
        if outChansSoFar > actualOutputChannels:
            if tileIndex == 0:
                descriptorOutputChannels = actualOutputChannels
            else:
                runningOutputChans = sum([outChans[0][1] for outChans in tiles[0:tileIndex]])
                descriptorOutputChannels = actualOutputChannels - runningOutputChans
        else:
            descriptorOutputChannels = subTiles[0][1]

        for subTileIndex, (maxW, maxN, mode) in enumerate(subTiles):

            lastSubTile = subTileIndex == len(subTiles) - 1
            genLastTile = lastTile and lastSubTile
            if lastSubTile:
                accum = 0
            else:
                accum = 1
            tapsOffset = (outputOffset * newInputDimZ) + (inputOffset * 8)
            fc = HwDescriptor(0, 0, 1, 0, mode, genTileIndex, genLastTile, stage.getName())
            fc.setupInputFullyConnected(inputOffset * inputStrideZ, accum, newInputDimZ, maxW)
            fc.setupOutputFullyConnected(outputOffset * outputStrideZ)
            fc.setupVectorsFullyConnected(0, tapsOffset * 2, maxN)

            # Store the actual output in the descriptor
            fc.actualOutChannels = descriptorOutputChannels

            if lastSubTile:
                postOp = [x for x in stage.getPostOp() if isinstance(x, ReLU)]
                if len(postOp) > 1:
                    raise ValueError("Only one post operation is supported. Found {}".format(postOp))
                elif len(postOp) == 1:
                    # General form of hardware Parametrized ReLU (PRelu):
                    #                     / a0*x, if x < t0
                    # f(x; t0, a0, a1) = |
                    #                     \ a1*x, if x >= t0
                    if hasattr(stage.post_op[0], 'negativeSlope') and stage.post_op[0].negativeSlope != 0:
                        post_param1 = stage.post_op[0].negativeSlope
                        # Check that the scaling with be 10 bits signed integer
                        assert(abs(int(round(1 / post_param1))) < 2 ** 9)
                        stage.setReLUScale(post_param1)
                        fc.setupRelu(0, 1, int(round(1 / post_param1)))
                        fc.setupScale(outputOffset * bytesPerPixel)
                    else:
                        fc.setupRelu(0, 0, 1)

            if stage.biasEnabled():
                fc.setupBias(outputOffset * bytesPerPixel)
            hwDescList.pushDescriptor(fc)
            genTileIndex += 1
            inputOffset += maxW
        outputOffset += maxN

    # Convert taps to hardware
    weights = stage.getWeights()
    taps = taps_hwck_to_cnnhw(np.transpose(weights.data, (0, 1, 3, 2)), newInputDimZ * splits, newOutputDimZ)

    #associate taps to the layer
    weights.data = taps
    weights.shape = (round_up(newOutputDimZ, 8), newInputDimZ * splits, weights.shape[0], weights.shape[1])

    concatOffset, finalOutputBufferSize = stage.getConcatParameters()

    if finalOutputBufferSize == 0:
        finalOutputBufferSize = outputBufferSize

    solution = ((inputStrideX, inputStrideY, inputStrideZ), (outputStrideX, outputStrideY, outputStrideZ),
                (newInputDimZ, newOutputDimZ, concatOffset, False, 0.0, 1.0, False, 0.0),
                (inputBufferSize, finalOutputBufferSize, tapsBufferSize, biasBufferSize), hwDescList, taps, None, None)

    return solution


def hwCompile(stage):

    if not stage.isHW:
        return None, None, None, None, None, None, None, None     # Fallback to SW

    if type(stage) in [HwConvolution, HwConvolutionPooling]:
        return compile_conv(stage)
    elif isinstance(stage, HwPooling):
        return compile_pooling(stage)
    elif isinstance(stage, HwFC):
        return compile_fc(stage)
    else:
        raise ValueError("Layer {} of type {} not support for HW compilation".format(stage.getName(), type(stage)))



def taps_hwck_to_cnnhw(data, new_c, new_k):
    def hwck_to_kchw(data):
        return np.transpose(data, (3, 2, 0, 1))

    fh, fw, c, k = data.shape

    # TODO::raise exception
    assert(k == new_k)

    #the way the taps need to be arranged for the HW block, is as if our outchannels%8=0
    #if this is not true, we will pad the taps with zeros
    padded_k =  ((k + 7) //8 ) *8

    newdata = np.zeros((padded_k, new_c, fh, fw), dtype=np.float16)

    newdata[0:k, 0:c, :, :] = hwck_to_kchw(data)
    newdata = newdata.reshape(padded_k, new_c * fh * fw).transpose()
    newdata = np.vstack(np.hsplit(newdata, padded_k // 8))
    return newdata.reshape((padded_k // 8, new_c, fh*fw, 8))

def regroup_taps(taps, splits):
    """ When doing split over the input channels,
        it is necessary to rearrange the taps (which
        have been converted for the hardware).

        input: output of `taps_hwck_to_cnnhw`
        splits: how many pieces is the input depth split into
    """

    # Reshape by adding an extra axis
    group, in_depth, ker_surf, out_ch = taps.shape

    taps = np.reshape(taps,
        (group, splits, in_depth // splits, ker_surf, out_ch))

    taps_stack = []
    for sp in range(splits):
        for g in range(group):
            taps_stack.append(taps[g, sp, :, :, :])

    return np.reshape(np.stack(taps_stack), (group, in_depth, ker_surf, out_ch))


def checkModeForValidity(iX, iY, iZ, oX, oY, oZ, kX, kY, kS, dataType, coeffType, mode):
    CNNHW_INTERNAL_MEMORY_SIZE = (128 * 1024)
    COEFF_PER_WORD_VALUES = {'FP16': 1, 'U8F': 2, '4BIT': 4, '2BIT': 8, '1BIT': 16, '1BITD': 16}
    BYTES_PER_PIXEL = {'FP16': 2, 'U8F': 1}
    blocks = 1 << mode
    inChansPerBlock = iZ // blocks
    coeffSetSize = kX*kY

    bytesPerLine = ((iX * BYTES_PER_PIXEL[dataType] + 15) // 16) * 16
    linesPerChannel = min(CNNHW_INTERNAL_MEMORY_SIZE // (bytesPerLine * iZ), iY, 2**9)

    coeffPerWord = COEFF_PER_WORD_VALUES[coeffType]
    coeffLPB = (inChansPerBlock * coeffSetSize + coeffPerWord - 1) // coeffPerWord

    if iX > 2**12 or iY > 2**12 or iZ > 2**11 or oZ > 2**11:
        return False, 0
    if kX > 2**4 or kY > 2**4 or kS > 2**4:
        return False, 0
    if inChansPerBlock > 2**11:
        return False, 0
    if coeffLPB > 2**8:
        return False, 0
    # if ((oX // kS) <= 4) and (linesPerChannel <= (kY + 2 * (kS + 1) + 1)):
    #     return False, 0
    # if ((oX // kS) > 4) and (linesPerChannel <= (kY + kS + 1 + 1)):
    #     print("linesPerChannel", linesPerChannel, kY, kS)
    #     return False, 0
    return True, (iZ // blocks) * kX * kY + [0, 5, 11, 19, 31][mode]

