# Copyright 2018 Intel Corporation.
# The source code, information and material ("Material") contained herein is
# owned by Intel Corporation or its suppliers or licensors, and title to such
# Material remains with Intel Corporation or its suppliers or licensors.
# The Material contains proprietary information of Intel or its suppliers and
# licensors. The Material is protected by worldwide copyright laws and treaty
# provisions.
# No part of the Material may be used, copied, reproduced, modified, published,
# uploaded, posted, transmitted, distributed or disclosed in any way without
# Intel's prior express written permission. No license under any patent,
# copyright or other intellectual property rights in the Material is granted to
# or conferred upon you, either expressly, by implication, inducement, estoppel
# or otherwise.
# Any license under such intellectual property rights must be express and
# approved by Intel in writing.

from .Layer import Layer
from Controllers.Tensor import UnpopulatedTensor, PopulatedTensor
import Models.Layouts as Layouts
from Controllers.TensorFormat import LayerIOFormat, LayerIOFormatPair
from Models.EnumDeclarations import PadStyle
from Controllers.Parsers.Parser.Upsampling import Upsampling
from Controllers.Parsers.Parser.crop import Crop
from Controllers.Parsers.Parser.Concat import Concat
from Controllers.Parsers.Parser.Pad import Pad
from Controllers.Parsers.Parser.Layer import MangledName, OriginalName
import numpy as np
import math

class Convolution2DUniversal(Layer):
    _supported_IO_formats = ()

    def __init__(self, *args):
        super().__init__(*args)
        self.padStyle = PadStyle.caffe

    def __repr__(self):
        repr = super().__repr__()
        repr += "\n  [Conv specific]:\n"
        repr += "      padStyle: %r\n" % (self.padStyle)
        repr += "      padding (t, b, l, r): (%r, %r, %r, %r)" % \
                    (self.pad_top, self.pad_bottom, self.pad_left, self.pad_right)
        return repr

    def loadKernelSize(self, kernelHeight, kernelWidth):
        self.kernelHeight = kernelHeight
        self.kernelWidth = kernelWidth

    def getKernelSize(self):
        return self.kernelHeight, self.kernelWidth

    def loadStrideSize(self, strideHeight, strideWidth):
        self.strideHeight = strideHeight
        self.strideWidth = strideWidth

    def getStrideSize(self):
        return (self.strideHeight, self.strideWidth)

    def loadPadding(self, pad_top, pad_bottom, pad_left, pad_right):
        self.pad_top    = pad_top
        self.pad_bottom = pad_bottom
        self.pad_left   = pad_left
        self.pad_right  = pad_right

    def setPadStyle(self, padStyle):
        if padStyle == 'VALID':
            self.padStyle = PadStyle.tfvalid
        elif padStyle == 'SAME':
            self.padStyle = PadStyle.tfsame
        elif padStyle == 'CAFFE':
            self.padStyle = PadStyle.caffe

    def force_symmetrical_padding(self):
        if self.padStyle == PadStyle.tfvalid or self.padStyle == PadStyle.tfsame:
            self.padStyle = PadStyle.caffe
            self.pad_top    = self.pad_bottom
            self.pad_left   = self.pad_right

    def getPadding(self):
        return (self.pad_top, self.pad_bottom, self.pad_left, self.pad_right)

    def loadDilation(self, dilationFactor):
        self.dilationFactor = dilationFactor

    def getDilation(self):
        return self.dilationFactor

    def setBiasEnabled(self, flag):
        self.hasBias = flag

    def biasEnabled(self):
        return self.hasBias

    def getBias(self):
        return self.bias

    def setBias(self, data):
        self.bias = PopulatedTensor(data)

    def getWeights(self):
        return self.weights

    def setWeights(self, data):
        # the weights data array is assumed to be in shape:
        #   (out_ch, in_ch, fH, fW)      - for plain Convolution
        #   (out_ch, group_size, fH, fW) - for GroupConvolution
        #       note: this is real group_size,
        #             not as self.groupSize, which refers to num_groups
        self.weights = PopulatedTensor(data)

    def loadOutputChannels(self, outputChannels):
        self.outputChannels = outputChannels

    def loadTrainedParameters(self, **kwargs):
        self.weights = PopulatedTensor(kwargs['weights'])
        try:
            self.bias = PopulatedTensor(kwargs['bias'])
        except:
            #print("No Bias")
            pass

class Deconvolution(Convolution2DUniversal):
    _supported_IO_formats = [
        LayerIOFormat.getLayerIOFormatPair(LayerIOFormatPair.NHWC_to_NHWC_W_stride)
        ]

    def __init__(self, *args):
        super().__init__(*args)

    def setNumGroups(self, num_groups):
        self.num_groups = num_groups

    def getNumGroups(self):
        return self.num_groups

    def getSplitGroupDeconvolution(self):
        """
            Generates a subgraph of Layers and Tensors that can replace a
            Deconvolution with group size parameter >= 1. This list consists of
            num_groups pairs of Crop + Deconv layers and a Concat layer that
            concats the output of the Deconv layers on the channel axis.

            return: List of layers: [Layer]
        """

        if self.num_groups == 1:
            return [self]

        channel_axis = 1
        assert(self.getInputTensorSizes()[0][channel_axis] % self.num_groups == 0)
        assert(self.outputChannels % self.num_groups == 0)
        in_group_size = self.getInputTensorSizes()[0][channel_axis] // self.num_groups
        out_group_size = self.outputChannels // self.num_groups

        in_group_shape = tuple([in_group_size if axis_i == channel_axis else axis
                               for axis_i, axis in enumerate(self.getInputTensorSizes()[0])])

        out_group_shape = tuple([out_group_size if axis_i == channel_axis else axis
                               for axis_i, axis in enumerate(self.getOutputTensorSizes()[0])])

        crop_group_layers = []
        deconv_group_layers = []
        name_suffix = self.name.stringifyOriginalName()
        for group_i in range(0, self.num_groups):
            group_i_crop_name = "crop_group_{0}_".format(group_i) + name_suffix
            group_i_crop_output = UnpopulatedTensor(in_group_shape)
            group_i_crop_output.setName(MangledName(OriginalName(group_i_crop_name)))
            group_i_crop = Crop(group_i_crop_name, [], [])
            group_i_crop.setInputTensorsAllFields(self.getInputTensors())
            group_i_crop.setOutputTensorsAllFields([group_i_crop_output])
            group_i_crop.setOffset(np.array([0, 0, in_group_size * group_i]))
            crop_group_layers.append(group_i_crop)

            group_i_deconv_name = "deconv_group_{0}_".format(group_i) + name_suffix
            group_i_deconv_output = UnpopulatedTensor(out_group_shape)
            group_i_deconv_output.setName(MangledName(OriginalName(group_i_deconv_name)))
            group_i_deconv = Deconvolution(
                group_i_deconv_name,
                group_i_crop.getOutputTensorNames(),
                group_i_deconv_name)
            group_i_deconv.setInputTensorsAllFields(group_i_crop.getOutputTensors())
            group_i_deconv.setOutputTensorsAllFields([group_i_deconv_output])
            group_i_deconv.loadPadding(*self.getPadding())
            group_i_deconv.padStyle = self.padStyle
            group_i_deconv.loadStrideSize(*self.getStrideSize())
            group_i_deconv.loadDilation(self.dilationFactor)
            group_i_deconv.setBiasEnabled(self.biasEnabled())
            group_i_deconv.loadKernelSize(*self.getKernelSize())
            group_i_deconv.setNumGroups(1)
            group_i_deconv.loadOutputChannels(out_group_size)
            if (self.biasEnabled()):
                group_i_deconv.setBias(self.bias.data[
                    :, :, :, group_i * out_group_size:(group_i + 1) * out_group_size])
                group_i_deconv.setBiasEnabled(True)
            group_i_deconv.setWeights(self.weights.data[
                :, group_i * in_group_size:(group_i + 1) * in_group_size, :, :])
            deconv_group_layers.append(group_i_deconv)

        concat_group = Concat("concat_group_deconv_" + name_suffix, [], [])
        concat_group.setInputTensorsAllFields(
            [tensor for layer in deconv_group_layers for tensor in layer.getOutputTensors()])
        concat_group.setOutputTensorsAllFields(self.getOutputTensors())
        concat_group.loadConcatAxis(channel_axis)

        return crop_group_layers + deconv_group_layers + [concat_group]

    def convert2Conv(self):
        # See http://deeplearning.net/software/theano/tutorial/conv_arithmetic.html#transposed-convolution

        if (self.strideHeight == 1 and self.strideWidth == 1 and
            self.pad_left == self.pad_right and
            self.pad_top == self.pad_bottom and
            self.pad_left == self.kernelWidth / 2 and
            self.pad_top == self.kernelHeight / 2):
            # Equivalent convolution has padding p' = k - p - 1
            pad_top = pad_bottom = self.kernelHeight - self.pad_top - 1
            pad_left = pad_right = self.kernelWidth - self.pad_left - 1

            conv = Convolution2D(self.name, self.inputTensorNames, self.outputTensorNames)
            for attr_name in self.__dict__:
                setattr(conv, attr_name, getattr(self, attr_name))

            conv.loadPadding(pad_top, pad_bottom, pad_left, pad_right)

            return {"conv" : conv, "upsampling" : None}
        elif self.pad_left == self.pad_right and \
             self.pad_top == self.pad_bottom:
            upsampled_tensor_name = MangledName(
                OriginalName(self.getInputTensorNames()[0].stringifyOriginalName()))
            upsampling_layer = Upsampling(
                self.name.stringifyOriginalName() + "_upsampling",
                self.getInputTensorNames(),
                [upsampled_tensor_name]
                )
            upsampling_layer.setInputTensors(self.getInputTensors())
            upsampling_layer.loadInputTensorSizes(self.getInputTensorSizes())
            upsampling_layer.compute_and_set_upsampling_parameters_for_deconv(
                self.strideWidth, self.strideHeight,
                self.kernelWidth, self.kernelHeight,
                self.pad_left, self.pad_top
                )

            upsampled_tensor = UnpopulatedTensor(
                shape=upsampling_layer.getOutputTensorSizes()[0])
            upsampled_tensor.setName(upsampling_layer.getOutputTensorNames()[0])
            upsampling_layer.setOutputTensors([upsampled_tensor])

            conv = Convolution2D(
                self.name.stringifyOriginalName() + "_as_conv",
                upsampling_layer.getOutputTensorNames(),
                self.outputTensorNames,
                )

            conv.setInputTensors([upsampled_tensor])
            conv.loadInputTensorSizes(upsampling_layer.getOutputTensorSizes())
            conv.setOutputTensors(self.getOutputTensors())
            conv.loadOutputTensorSizes(self.getOutputTensorSizes())
            conv.loadPadding(0, 0, 0, 0)
            conv.loadStrideSize(1, 1)
            conv.loadDilation(self.dilationFactor)
            conv.setBiasEnabled(self.biasEnabled())
            conv.loadKernelSize(*self.getKernelSize())
            conv.loadGroupSize(self.num_groups)
            conv.loadOutputChannels(self.outputChannels)
            if (self.biasEnabled()):
                conv.bias = self.bias
            conv.weights = self.weights

            return {"conv" : conv, "upsampling" : upsampling_layer}

class Convolution2D(Convolution2DUniversal):
    _supported_IO_formats = [
        LayerIOFormat.getLayerIOFormatPair(LayerIOFormatPair.NHWC_to_NHWC_W_stride)
        ]
    def __init__(self, *args):
        super().__init__(*args)

        # This is a hack in order to force certain Layouts for the generated group
        # conv.
        self.from_depthwise = False

        # This is a hack for the profiler fix in new parser
        self.isDepthwise=False

    def loadGroupSize(self, groupSize):
        self.groupSize = groupSize

    def getGroupSize(self):
        return self.groupSize

    def genGroupConvs(self):

        group_filter = (1, self.getGroupSize(), 1, 1)
        group_shape = list(map(lambda a, b: a // b, self.getInputTensors()[0].getShape(), group_filter))
        group_channels = group_shape[1]

        out_ch = self.getOutputTensors()[0].getShape()[1]
        output_group_shape = list(map(lambda a, b: a // b, self.getOutputTensors()[0].getShape(), group_filter))

        group_convs = []

        for group in range(self.getGroupSize()):

            # Take our slice of Weights
            group_N_weights = self.weights.data[self.weights.data.shape[0] // self.getGroupSize() * group:
                                self.weights.data.shape[0] // self.getGroupSize() * (group + 1), ]

            # Our slice of Input...
            group_N_input = UnpopulatedTensor(group_shape)
            group_N_input.setName(MangledName(OriginalName(self.getInputTensors()[0].name.stringifyOriginalName())))


            # And Output...
            outputSlice = UnpopulatedTensor((output_group_shape))
            outputSlice.setName(MangledName(OriginalName(self.getOutputTensors()[0].name.stringifyOriginalName())))


            # Create the new Convolution representhing this group.
            new_name = self.name.stringifyOriginalName()
            l = Convolution2D(new_name, None, None)

            l.setInputTensorsAllFields([group_N_input])
            l.setOutputTensorsAllFields([outputSlice])

            # Set group size
            l.loadGroupSize(1)
            l.sliced = (group, self.getGroupSize())

            # Apply our split of Weights
            l.setWeights(group_N_weights)

            # Copy some of the other properties of the original convolution.
            l.loadKernelSize(*self.getKernelSize())
            l.loadStrideSize(*self.getStrideSize())
            l.loadPadding(*self.getPadding())
            l.padStyle = self.padStyle
            l.loadDilation(self.getDilation())

            if self.biasEnabled():
                bias_data = self.getBias().data.flatten()
                l.setBias(bias_data[group * (out_ch // self.getGroupSize()) : (group + 1) * (out_ch // self.getGroupSize())])  # Correct?
                l.setBiasEnabled(True)
            else:
                l.setBiasEnabled(False)

            c = Crop("crop_" + l.name.stringifyOriginalName(), [], [])
            c.setInputTensorsAllFields(self.getInputTensors())
            c.setOutputTensorsAllFields([group_N_input])
            c.setOffset(np.array([self.getInputTensors()[0].shape[1] // self.getGroupSize() * group, 0, 0]))
            c.setImplicit()

            # This is a hack in order to keep the same performance for the
            # CaffeMobileNets and Alexnet. This hack will force the generated squence
            # of Crop + Conv + Concate to operate in certain layouts for these 2 cases.
            # This might affect different usecase similar to the ones encountered in
            # the above in diffrent ways.
            # The forcing below of the io formats should be removed once the NCSDK-1276
            if self.from_depthwise:
                c.setSelectedIOFormats(
                    [io_fmt_pair for io_fmt_pair in c.getSelectedIOFormats()
                     if all([fmt.layout != Layouts.NHCW for tname, fmt in io_fmt_pair.items()])])

            group_convs.append([l,c])

        return group_convs

    def getSplitGroupConvolutionForShaves(self):
        """
            Generates a subgraph of Layers and Tensors that can replace a
            Convolution with group size parameter >= 1. This list consists of
            num_groups pairs of Crop + Conv layers and a Concat layer that
            concats the output of the Conv layers on the channel axis.

            return: List of layers: [Layer]
        """
        # getGroupSize() returns the number of groups not the size of the group.
        num_groups = self.getGroupSize()

        if num_groups == 1:
            return [self]

        assert(self.getInputTensorSizes()[0][Layouts.Axes.C] % num_groups == 0)
        assert(self.outputChannels % num_groups == 0)
        in_group_size = self.getInputTensorSizes()[0][Layouts.Axes.C] // num_groups
        out_group_size = self.outputChannels // num_groups

        in_group_shape = tuple([in_group_size if axis_i == Layouts.Axes.C else axis
                               for axis_i, axis in enumerate(self.getInputTensorSizes()[0])])

        out_group_shape = tuple([out_group_size if axis_i == Layouts.Axes.C else axis
                               for axis_i, axis in enumerate(self.getOutputTensorSizes()[0])])

        crop_group_layers = []
        conv_group_layers = []
        name_suffix = self.name.stringifyOriginalName()
        for group_i in range(0, num_groups):
            group_i_crop_name = "crop_group_{0}_".format(group_i) + name_suffix
            group_i_crop_output = UnpopulatedTensor(in_group_shape)
            group_i_crop_output.setName(MangledName(OriginalName(group_i_crop_name)))
            group_i_crop = Crop(group_i_crop_name, [], [])
            group_i_crop.setInputTensorsAllFields(self.getInputTensors())
            group_i_crop.setOutputTensorsAllFields([group_i_crop_output])
            group_i_crop.setOffset(np.array([0, 0, in_group_size * group_i]))
            crop_group_layers.append(group_i_crop)

            group_i_conv_name = "conv_group_{0}_".format(group_i) + name_suffix
            group_i_conv_output = UnpopulatedTensor(out_group_shape)
            group_i_conv_output.setName(MangledName(OriginalName(group_i_conv_name)))
            group_i_conv = Convolution2D(
                group_i_conv_name,
                group_i_crop.getOutputTensorNames(),
                group_i_conv_name)
            group_i_conv.setInputTensorsAllFields(group_i_crop.getOutputTensors())
            group_i_conv.setOutputTensorsAllFields([group_i_conv_output])
            group_i_conv.loadPadding(*self.getPadding())
            group_i_conv.loadStrideSize(*self.getStrideSize())
            group_i_conv.padStyle = self.padStyle
            group_i_conv.loadDilation(self.dilationFactor)
            group_i_conv.setBiasEnabled(self.biasEnabled())
            group_i_conv.loadKernelSize(*self.getKernelSize())
            group_i_conv.loadGroupSize(1)
            group_i_conv.loadOutputChannels(out_group_size)
            if (self.biasEnabled()):
                group_i_conv.setBias(self.bias.data[
                    :, :, :, group_i * out_group_size:(group_i + 1) * out_group_size])
                group_i_conv.setBiasEnabled(True)
            group_i_conv.setWeights(self.weights.data[
                group_i * out_group_size:(group_i + 1) * out_group_size, :, :, :])
            conv_group_layers.append(group_i_conv)

        concat_group = Concat("concat_group_conv_" + name_suffix, [], [])
        concat_group.setInputTensorsAllFields(
            [tensor for layer in conv_group_layers for tensor in layer.getOutputTensors()])
        concat_group.setOutputTensorsAllFields(self.getOutputTensors())
        concat_group.loadConcatAxis(Layouts.Axes.C)

        return crop_group_layers + conv_group_layers + [concat_group]

    def implementAsPadPlusConv(self):
        name_suffix = self.name.stringifyOriginalName()
        (pad_t, pad_b, pad_l, pad_r) = self.getPadding()
        (kh, kw) = self.getKernelSize()

        # if the (conv) layer is already suitable for HW implementaton,
        # then skip adding the Pad layer
        if (pad_l == 0 or pad_l == (kw // 2)) and \
        (pad_t == 0 or pad_t == (kh // 2)) and \
        (pad_t == pad_b) and \
        (pad_l == pad_r):
            return [self]

        padded_tensor_name = MangledName(
                OriginalName("padded_" + self.getInputTensorNames()[0].stringifyOriginalName()))

        pad_layer_name = "pad_" + name_suffix
        pad_layer = Pad(pad_layer_name,
                    self.getInputTensorNames(),
                    [padded_tensor_name])
        pad_layer.set_padding_size(*self.getPadding())
        pad_layer.setInputTensorsAllFields(self.getInputTensors())
        pad_layer.loadInputTensorSizes(self.getInputTensorSizes())
        pad_layer.compute_output_shape()

        padded_tensor = UnpopulatedTensor(
            shape=pad_layer.getOutputTensorSizes()[0])
        padded_tensor.setName(padded_tensor_name)
        pad_layer.setOutputTensors([padded_tensor])
        # pad_layer.setOutputTensorsAllFields([pad_layer_name])

        self.loadPadding(0, 0, 0, 0)
        self.padStyle = PadStyle.caffe
        self.setInputTensorsAllFields(pad_layer.getOutputTensors())
        self.loadInputTensorSizes(pad_layer.getOutputTensorSizes())

        return [pad_layer, self]

class ConvolutionDepthWise2D(Convolution2DUniversal):
    _supported_IO_formats = [
        LayerIOFormat.getLayerIOFormatPair(LayerIOFormatPair.NHWC_to_NHWC_W_stride),
        LayerIOFormat.getLayerIOFormatPair(LayerIOFormatPair.NHCW_to_NHCW_C_stride)
        ]

    def __init__(self, *args):
        super().__init__(*args)

        # This is a hack for the profiler fix in new parser
        # In mobilenet ssd, priorbox forces convertion layers. If type 'any' of TensorFormat
        # is implemented, this should not be necessary
        self.isDepthwise = True

    def loadGroupSize(self, groupSize):
        # TODO: self.groupSize actually refers to number of groups
        #       A cleanup should be done here to fix this confusion
        self.groupSize = groupSize

    def getGroupSize(self):
        return self.groupSize

    def convert2GroupConv(self, scheduler):

        group_conv = Convolution2D('group_' + self.name.stringifyOriginalName(), [], [])

        # Copy attributes
        for attr_name in self.__dict__:
            setattr(group_conv, attr_name, getattr(self, attr_name))

        implementAsSwLayer, new_group_size = self.getNewGroupSize(scheduler)
        if implementAsSwLayer:
            return self

        new_weights = np.zeros((self.weights.data.shape[0], self.groupSize*self.weights.data.shape[1]//new_group_size, self.kernelHeight, self.kernelWidth))

        stepk = self.weights.shape[0]//self.groupSize
        stepc = self.weights.shape[1]
        max_c = new_weights.shape[1]
        for idx, group in enumerate(range(self.groupSize)):
            group_N_weights = self.weights.data[self.weights.data.shape[0] // self.groupSize * group:
                                self.weights.data.shape[0] // self.groupSize * (group + 1), ]
            new_weights[stepk*idx:stepk*(idx+1),  stepc*(idx%max_c):stepc*((idx%max_c)+1), ...] = group_N_weights

        group_conv.setWeights(new_weights)
        group_conv.groupSize = new_group_size
        # This is a hack in order to force certain Layouts for the generated group
        # conv.
        group_conv.from_depthwise = True

        return group_conv

    def getNewGroupSize(self, scheduler):

        i = self.getInputTensors()[0].shape
        o = self.getOutputTensors()[0].shape
        # Ensure compatibility with old code
        stage = {'kx': self.kernelWidth, 'ky': self.kernelHeight,
                 'sx': self.strideWidth, 'sy': self.strideHeight,
                 'ic': i[1], 'oc': o[1],
                 'is': [i[3], i[2]],
                 'os': [o[3], o[2]],
                 'pkx':self.kernelWidth, 'pky':self.kernelHeight,
                 'pkx':self.strideWidth, 'pky':self.strideHeight}
        sw, newGroups = scheduler.optimize_depthwise_convolution(stage = stage)

        return sw, newGroups
